# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for the 'zhmc cpc hw-message' command group.
"""

import random
import urllib3
import pytest
import zhmcclient
# pylint: disable=unused-import
from zhmcclient.testutils import hmc_definition  # noqa: F401
from .utils import zhmc_session  # noqa: F401
# pylint: enable=unused-import

from .utils import run_zhmc
from .test_console_hwmessage import assert_messages

urllib3.disable_warnings()


# pylint: disable=inconsistent-return-statements
def pick_cpc_with_hwmessages(cpcs):
    """
    Pick a CPC that has at least one hardware message.
    """
    for cpc in cpcs:
        messages = cpc.hw_messages.list()
        if messages:
            return cpc, messages

    pytest.skip("Did not find a CPC with hardware messages")


# Expected output properties when no other options are specified
CPC_HWMSG_PROPS_DEFAULT = [
    "timestamp-utc",
    "message-id",
    "text",
]

# Expected output properties when --all option is specified
# Note: These are the z16 properties. Additional properties are allowed.
CPC_HWMSG_PROPS_ALL = [
    "timestamp-utc",
    "message-id",
    "text",
    "class",
    "details",
    "element-id",
    "element-uri",
    "parent",
    "service-supported",
    "text",
    "timestamp",
]

# Artificial output properties (that are not properties on the HardwareMessage
# object).
CPC_HWMSG_PROPS_ARTIFICIAL = [
    "timestamp-utc",
    "message-id",
]


TESTCASES_CPC_HWMSG_LIST = [
    # Testcases for test_cpc_hwmsg_list()
    # Each list item is a testcase, which is a tuple with items:
    # - options (list of str): Options for 'zhmc cpc list' command
    # - exp_props (list of str): Expected property names in output
    (
        [],
        CPC_HWMSG_PROPS_DEFAULT,
    ),
    (
        ['--all'],
        CPC_HWMSG_PROPS_ALL,
    ),
]


@pytest.mark.parametrize(
    "options, exp_props",
    TESTCASES_CPC_HWMSG_LIST
)
@pytest.mark.parametrize(
    "out_format",
    ['json', 'csv']
)
def test_cpc_hwmsg_list(
        zhmc_session, out_format, options, exp_props):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test 'cpc hw-message list' command for output formats json and csv.
    """
    client = zhmcclient.Client(zhmc_session)
    cpcs = client.cpcs.list()
    cpc, messages = pick_cpc_with_hwmessages(cpcs)

    # Reduce messages to be listed since --all takes a long time for many
    # messages
    num_messages = min(5, len(messages))
    messages = messages[-num_messages:]
    first_ts_dt = zhmcclient.datetime_from_timestamp(
        messages[0].prop('timestamp'))
    first_ts_str = first_ts_dt.strftime('%Y-%m-%d %H:%M:%S.%f')

    args = ['-o', out_format, 'cpc', 'hw-message', 'list', '--begin',
            first_ts_str, cpc.name] + options
    rc, stdout, stderr = run_zhmc(args)

    assert_messages(rc, stdout, stderr, out_format, 'list', 0, "",
                    exp_props, messages)


@pytest.mark.parametrize(
    "out_format",
    ['json', 'csv']
)
def test_cpc_hwmsg_show(zhmc_session, out_format):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test 'cpc hw-message show' command.
    """
    client = zhmcclient.Client(zhmc_session)
    cpcs = client.cpcs.list()
    cpc, messages = pick_cpc_with_hwmessages(cpcs)

    test_message = random.choice(messages)  # nosec
    test_message_id = test_message.prop('element-id')

    rc, stdout, stderr = run_zhmc(
        ['-o', out_format, 'cpc', 'hw-message', 'show', cpc.name,
         test_message_id])

    assert_messages(rc, stdout, stderr, out_format, 'show', 0, "",
                    CPC_HWMSG_PROPS_DEFAULT, [test_message])


# Note: We do not have end2end tests for modifying commands:
# 'cpc hw-message delete'
# 'cpc hw-message get-service-info'
# 'cpc hw-message request-service'
# 'cpc hw-message decline-service'



# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Function tests for _session_file.py module.
"""

import os
import re
from datetime import datetime, timezone
import tempfile
import pytest
from zhmcclient_mock import FakedSession

from zhmccli._session_file import HMCSession, HMCSessionFile, \
    HMCSessionException, HMCSessionNotFound, HMCSessionAlreadyExists, \
    HMCSessionFileNotFound, HMCSessionFileError, HMCSessionFileFormatError, \
    BLANKED_OUT_STRING, TIME_FORMAT, now_str


def create_session_file(sf_str):
    """
    Create an HMC session file with the specified content.

    The content string does not need to be valid according to the HMC session
    file schema, in order to be able to provoke format errors.
    The HMC session file is closed upon return.

    If the 'sf_str' parameter is None, no session file is created, and the
    path name of a non-existing file is returned.

    Parameters:
      sf_str (str): Content. May be None.

    Returns:
      HMCSessionFile: HMC session file.
    """
    home_dir = os.path.expanduser('~')
    # pylint: disable=consider-using-with
    file = tempfile.NamedTemporaryFile(
        mode='w', encoding="utf-8", delete=False,
        suffix='yaml', prefix='.test_zhmc_sessions_', dir=home_dir)
    filepath = file.name
    if sf_str is None:
        file.close()
        os.remove(filepath)
    else:
        file.write(sf_str)
        file.close()
    return HMCSessionFile(filepath)


def delete_session_file(session_file):
    """
    Delete an HMC session file, if it exists.

    Parameters:
      session_file (HMCSessionFile): HMC session file.
    """
    filepath = session_file.filepath
    if os.path.exists(filepath):
        os.remove(filepath)


def session_file_content(session_file):
    """
    Return the content of an HMC session file, if it exists.

    If it does not exist, return None.

    Parameters:
      session_file (HMCSessionFile): HMC session file.

    Returns:
      str: Content of HMC session file, or None.
    """
    filepath = session_file.filepath
    if os.path.exists(filepath):
        with open(filepath, "r", encoding="utf-8") as fp:
            sf_str = fp.read()
        return sf_str
    return None


def assert_session_equal(act_session, exp_session, session_name):
    """
    Assert that two HMCSession objects are equal.
    """
    assert isinstance(act_session, HMCSession)
    assert isinstance(exp_session, HMCSession)
    for attr in ("host", "userid", "session_id", "ca_verify", "ca_cert_path",):
        act_value = getattr(act_session, attr)
        exp_value = getattr(exp_session, attr)
        assert act_value == exp_value, (
            f"Unexpected value of HMCSession {session_name!r} attribute "
            f"{attr!r}:\n"
            f"  Actual value: {act_value!r}\n"
            f"  Expected value: {exp_value!r}")
    assert_time_str(act_session.creation_time, exp_session.creation_time, 5)


def assert_session_file_content(session_file, exp_sf_str):
    """
    Assert that the HMC session file has the expected content.

    Because loading an HMC session file changes a non-existing file to exist
    with content '{}', this needs to be considered (they are considered
    equal).
    """
    act_sf_str = session_file_content(session_file)
    if exp_sf_str is None and act_sf_str == "{}":
        return
    assert act_sf_str == exp_sf_str, (
        f"HMC session file {session_file.filepath!r} has unexpected content:\n"
        f"  Actual content:\n{act_sf_str}\n"
        f"  Expected content:\n{exp_sf_str}\n")


@pytest.mark.parametrize(
    "exc_type",
    [
        HMCSessionNotFound,
        HMCSessionAlreadyExists,
        HMCSessionFileNotFound,
        HMCSessionFileError,
        HMCSessionFileFormatError
    ]
)
def test_session_file_exception(exc_type):
    """
    Test session file exception.
    """
    message = "Message"

    # The code to be tested
    exc = exc_type(message)

    assert isinstance(exc, Exception)
    assert isinstance(exc, HMCSessionException)
    assert str(exc) == message
    assert exc.args[0] == message


def test_session_repr():
    """
    Test HMCSession.__repr__().
    """
    host = "my_host"
    userid = "my_userid"
    session_id = "my_session_id"
    ca_verify = False
    ca_cert_path = None

    session = HMCSession(
        host=host,
        userid=userid,
        session_id=session_id,
        ca_verify=ca_verify,
        ca_cert_path=ca_cert_path)

    # The code to be tested
    repr_str = repr(session)

    exp_repr_pattern = (
        re.escape(
            "HMCSession("
            f"host={host!r}, "
            f"userid={userid!r}, "
            f"session_id={BLANKED_OUT_STRING}, "
            f"ca_verify={ca_verify!r}, "
            f"ca_cert_path={ca_cert_path!r}, "
            f"creation_time=") + r"'(.*)'\)")

    m = re.match(exp_repr_pattern, repr_str)
    assert m is not None, (
        "Unexpected repr string:\n"
        f"Actual: {repr_str}\n"
        f"Expected pattern: {exp_repr_pattern}")

    # Verify creation_time value
    creation_time = m.group(1)
    assert_time_str(creation_time, datetime.utcnow(), 5)


def assert_time_str(time_str, exp_time, delta):
    """
    Assert for time_str:
    * that it has the required format.
    * That it is within a delta of the expected point in time.

    Parameters:
      time_str (str): The point in time to be checked.
      exp_time (datetime or str): The expected point in time.
      delta (int): The allowed time delta, in seconds
    """
    if isinstance(exp_time, str):
        exp_time = datetime.strptime(exp_time, TIME_FORMAT)
        exp_time.replace(tzinfo=timezone.utc)
    assert isinstance(exp_time, datetime)

    try:
        time_dt = datetime.strptime(time_str, TIME_FORMAT)
    except ValueError:
        raise AssertionError(
            f"Time string does not have the required format: {time_str!r}")
    time_dt.replace(tzinfo=timezone.utc)

    assert abs((exp_time - time_dt).total_seconds()) <= delta, (
        f"Unexpected time: {time_str} - expected was {exp_time} with a "
        f"delta of {delta} s")


def test_session_as_dict():
    """
    Test HMCSession.as_dict().
    """
    host = "my_host"
    userid = "my_userid"
    session_id = "my_session_id"
    ca_verify = False
    ca_cert_path = None

    session = HMCSession(
        host=host,
        userid=userid,
        session_id=session_id,
        ca_verify=ca_verify,
        ca_cert_path=ca_cert_path)

    # The code to be tested
    session_dict = session.as_dict()

    assert isinstance(session_dict, dict)
    assert len(session_dict) == 6
    assert session_dict['host'] == host
    assert session_dict['userid'] == userid
    assert session_dict['session_id'] == session_id
    assert session_dict['ca_verify'] == ca_verify
    assert session_dict['ca_cert_path'] == ca_cert_path
    assert_time_str(session_dict['creation_time'], datetime.utcnow(), 5)


TESTCASES_SESSION_FROM_ZHMCCLIENT = [

    # Each item is a testcase for testing HMCSession.from_zhmcclient_session(),
    # with the following items:
    #
    # * verify_cert (bool/str): Input parameter for zhmcclient session.
    # * exp_ca_verify (bool): Expected property from function result.
    # * exp_ca_cert_path (str): Expected property from function result.

    (
        False,
        False,
        None
    ),
    (
        True,
        True,
        None
    ),
    (
        'dir/file.pem',
        True,
        'dir/file.pem'
    ),
]


@pytest.mark.parametrize(
    "verify_cert, exp_ca_verify, exp_ca_cert_path",
    TESTCASES_SESSION_FROM_ZHMCCLIENT
)
def test_session_from_zhmcclient(verify_cert, exp_ca_verify, exp_ca_cert_path):
    """
    Test HMCSession.from_zhmcclient_session().
    """
    host = "my_host"
    userid = "my_userid"
    session_id = "my_session_id"

    faked_session = FakedSession(host, 'hmc1', '2.16', '4.10', userid=userid)
    faked_session._session_id = session_id  # pylint: disable=protected-access
    faked_session._verify_cert = verify_cert  # pylint: disable=protected-access

    # The code to be tested
    hmc_session = HMCSession.from_zhmcclient_session(faked_session)

    assert hmc_session.host == host
    assert hmc_session.userid == userid
    assert hmc_session.session_id == session_id
    assert hmc_session.ca_verify == exp_ca_verify
    assert hmc_session.ca_cert_path == exp_ca_cert_path
    assert_time_str(hmc_session.creation_time, datetime.utcnow(), 5)


TESTCASES_SESSION_FILE_REPR = [

    # Each item is a testcase for testing HMCSessionFile.__repr__(),
    # with the following items:
    #
    # * sf_str (str/None): Content for HMC session file, or None for an empty
    #   HMC session file.
    # * exp_repr_data (str): Expected data shown in repr string.

    (
        None,
        "{}"
    ),
    (
        """
default:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:14:58"
""",
        "{'default': {...}}"
    ),
]


@pytest.mark.parametrize(
    "sf_str, exp_repr_data",
    TESTCASES_SESSION_FILE_REPR
)
def test_session_file_repr(sf_str, exp_repr_data):
    """
    Test HMCSessionFile.__repr__().
    """
    session_file = create_session_file(sf_str)
    filepath = session_file.filepath

    try:

        # The code to be tested
        repr_str = repr(session_file)

        # pylint: disable=protected-access
        exp_repr_str = (
            "HMCSessionFile("
            f"filepath={filepath!r}, "
            f"data={exp_repr_data})")

        assert repr_str == exp_repr_str

    finally:
        delete_session_file(session_file)


TESTCASES_SESSION_FILE_LIST = [

    # Each item is a testcase for testing HMCSessionFile.list(), with the
    # following items:
    #
    # * desc (str): Testcase description.
    # * initial_content (str): Content of initial HMC session file that is
    #   created before calling the function to be tested. None means there is
    #   no initial HMC session file.
    # * exp_result (dict): Expected result from the function to be tested.
    #   Key(str): Session name, Value(dict): HMC session kwargs.
    # * exp_exc_type (exc): Expected exception class. None for no exception.
    # * exp_exc_msg (str): Expected pattern for exception message. None for
    #   no exception.

    (
        "no session file",
        None,
        {},
        None, None
    ),
    (
        "invalid YAML format",
        "{\n",
        None,
        HMCSessionFileFormatError,
        "Invalid YAML syntax in HMC session file"
    ),
    (
        "invalid schema",
        "42\n",
        None,
        HMCSessionFileFormatError,
        "Invalid data format in HMC session file"
    ),
    (
        "empty session file",
        "{}\n",
        {},
        None, None
    ),
    (
        "session file with one valid session",
        """
default:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:14:58"
        """,
        {
            'default': dict(
                host="my_host",
                userid="my_userid",
                session_id="my_session_id",
                ca_verify=False,
                ca_cert_path=None,
                creation_time="2025-05-06 07:14:58")
        },
        None, None
    ),
    (
        "session file with two valid sessions",
        """
default:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:14:58"
s2:
  host: my_host2
  userid: my_userid2
  session_id: my_session_id2
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        {
            'default': dict(
                host="my_host",
                userid="my_userid",
                session_id="my_session_id",
                ca_verify=False,
                ca_cert_path=None,
                creation_time="2025-05-06 07:14:58"),
            's2': dict(
                host="my_host2",
                userid="my_userid2",
                session_id="my_session_id2",
                ca_verify=False,
                ca_cert_path=None,
                creation_time="2025-05-06 07:20:58")
        },
        None, None
    ),
    (
        "session file with one valid session with missing properties",
        """
default:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
        """,
        None,
        HMCSessionFileFormatError,
        "Invalid data format in HMC session file"
    ),
]


@pytest.mark.parametrize(
    "desc, initial_content, exp_result, exp_exc_type, exp_exc_msg",
    TESTCASES_SESSION_FILE_LIST
)
def test_session_file_list(
        desc, initial_content, exp_result, exp_exc_type, exp_exc_msg):
    # pylint: disable=unused-argument
    """
    Test HMCSessionFile.list().
    """
    session_file = create_session_file(initial_content)

    try:
        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                session_file.list()

            exc = exc_info.value
            msg = str(exc)
            m = re.match(exp_exc_msg, msg)
            assert m, \
                "Unexpected exception message:\n" \
                "  expected pattern: {!r}\n" \
                "  actual message: {!r}".format(exp_exc_msg, msg)
        else:

            # The code to be tested
            result = session_file.list()

            assert isinstance(result, dict)
            assert set(result.keys()) == set(exp_result.keys())
            for name, act_session in result.items():
                assert name in exp_result
                exp_session = HMCSession(**exp_result[name])
                assert_session_equal(act_session, exp_session, name)

        assert_session_file_content(session_file, initial_content)

    finally:
        delete_session_file(session_file)


TESTCASES_SESSION_FILE_GET = [

    # Each item is a testcase for testing HMCSessionFile.get(), with the
    # following items:
    #
    # * desc (str): Testcase description.
    # * initial_content (str): Content of initial HMC session file that is
    #   created before calling the function to be tested. None means there is
    #   no initial HMC session file.
    # * session_name (str): Session name to be retrieved.
    # * exp_session_kwargs (dict): Expected HMC session, or None for exception.
    # * exp_exc_type (exc): Expected exception class. None for no exception.
    # * exp_exc_msg (str): Expected pattern for exception message. None for
    #   no exception.

    (
        "no session file",
        None,
        "foo",
        None,
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
    (
        "invalid YAML format",
        "{\n",
        "foo",
        None,
        HMCSessionFileFormatError,
        "Invalid YAML syntax in HMC session file"
    ),
    (
        "invalid schema",
        "42\n",
        "foo",
        None,
        HMCSessionFileFormatError,
        "Invalid data format in HMC session file"
    ),
    (
        "empty session file",
        "{}\n",
        "foo",
        None,
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
    (
        "session file with two valid sessions, one of those requested",
        """
default:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:18:58"
s2:
  host: my_host2
  userid: my_userid2
  session_id: my_session_id2
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "s2",
        dict(
            host="my_host2",
            userid="my_userid2",
            session_id="my_session_id2",
            ca_verify=False,
            ca_cert_path=None,
            creation_time="2025-05-06 07:20:58",
        ),
        None, None
    ),
    (
        "session file with two valid sessions, none of those requested",
        """
default:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:18:58"
s2:
  host: my_host2
  userid: my_userid2
  session_id: my_session_id2
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "foo",
        None,
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
]


@pytest.mark.parametrize(
    "desc, initial_content, session_name, exp_session_kwargs, exp_exc_type, "
    "exp_exc_msg",
    TESTCASES_SESSION_FILE_GET
)
def test_session_file_get(
        desc, initial_content, session_name, exp_session_kwargs, exp_exc_type,
        exp_exc_msg):
    # pylint: disable=unused-argument
    """
    Test HMCSessionFile.get().
    """
    session_file = create_session_file(initial_content)

    try:
        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                session_file.get(session_name)

            exc = exc_info.value
            msg = str(exc)
            m = re.match(exp_exc_msg, msg)
            assert m, \
                "Unexpected exception message:\n" \
                "  expected pattern: {!r}\n" \
                "  actual message: {!r}".format(exp_exc_msg, msg)
        else:

            # The code to be tested
            session = session_file.get(session_name)
            exp_session = HMCSession(**exp_session_kwargs)

            assert isinstance(session, HMCSession)
            assert_session_equal(session, exp_session, session_name)

        assert_session_file_content(session_file, initial_content)

    finally:
        delete_session_file(session_file)


TESTCASES_SESSION_FILE_ADD = [

    # Each item is a testcase for testing HMCSessionFile.add(), with the
    # following items:
    #
    # * desc (str): Testcase description.
    # * initial_content (str): Content of initial HMC session file that is
    #   created before calling the function to be tested. None means there is
    #   no initial HMC session file.
    # * session_name (str): Session name to be added.
    # * session_kwargs (dict): Input args for session to be added.
    # * exp_exc_type (exc): Expected exception class. None for no exception.
    # * exp_exc_msg (str): Expected pattern for exception message. None for
    #   no exception.

    (
        "no session file",
        None,
        "foo",
        dict(
            host="my_host",
            userid="my_userid",
            session_id="my_session_id",
            ca_verify=False,
            ca_cert_path=None,
        ),
        None, None
    ),
    (
        "empty session file",
        "{}\n",
        "foo",
        dict(
            host="my_host",
            userid="my_userid",
            session_id="my_session_id",
            ca_verify=False,
            ca_cert_path=None,
        ),
        None, None
    ),
    (
        "session file with one valid session, new session does not exist",
        """
default:
  host: my_host_def
  userid: my_userid_def
  session_id: my_session_id_def
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:18:58"
        """,
        "foo",
        dict(
            host="my_host",
            userid="my_userid",
            session_id="my_session_id",
            ca_verify=False,
            ca_cert_path=None,
        ),
        None, None
    ),
    (
        "session file with one valid session, new session exists",
        """
foo:
  host: my_host_def
  userid: my_userid_def
  session_id: my_session_id_def
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "foo",
        dict(
            host="my_host",
            userid="my_userid",
            session_id="my_session_id",
            ca_verify=False,
            ca_cert_path=None,
        ),
        HMCSessionAlreadyExists,
        "Session already exists in HMC session file"
    ),
]


@pytest.mark.parametrize(
    "desc, initial_content, session_name, session_kwargs, exp_exc_type, "
    "exp_exc_msg",
    TESTCASES_SESSION_FILE_ADD
)
def test_session_file_add(
        desc, initial_content, session_name, session_kwargs, exp_exc_type,
        exp_exc_msg):
    # pylint: disable=unused-argument
    """
    Test HMCSessionFile.add().
    """
    session_file = create_session_file(initial_content)
    session = HMCSession(**session_kwargs)

    try:
        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                session_file.add(session_name, session)

            exc = exc_info.value
            msg = str(exc)
            m = re.match(exp_exc_msg, msg)
            assert m, \
                "Unexpected exception message:\n" \
                "  expected pattern: {!r}\n" \
                "  actual message: {!r}".format(exp_exc_msg, msg)

            assert_session_file_content(session_file, initial_content)

        else:

            # The code to be tested
            session_file.add(session_name, session)

            retrieved_session = session_file.get(session_name)

            assert_session_equal(session, retrieved_session, session_name)

    finally:
        delete_session_file(session_file)


TESTCASES_SESSION_FILE_REMOVE = [

    # Each item is a testcase for testing HMCSessionFile.remove(), with the
    # following items:
    #
    # * desc (str): Testcase description.
    # * initial_content (str): Content of initial HMC session file that is
    #   created before calling the function to be tested. None means there is
    #   no initial HMC session file.
    # * session_name (str): Session name to be removed.
    # * exp_exc_type (exc): Expected exception class. None for no exception.
    # * exp_exc_msg (str): Expected pattern for exception message. None for
    #   no exception.

    (
        "no session file",
        None,
        "foo",
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
    (
        "empty session file",
        "{}\n",
        "foo",
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
    (
        "session file with one valid session that is session to be removed",
        """
foo:
  host: my_host_def
  userid: my_userid_def
  session_id: my_session_id_def
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "foo",
        None, None
    ),
    (
        "session file with one valid session that is not session to be removed",
        """
default:
  host: my_host_def
  userid: my_userid_def
  session_id: my_session_id_def
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "foo",
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
]


@pytest.mark.parametrize(
    "desc, initial_content, session_name, exp_exc_type, exp_exc_msg",
    TESTCASES_SESSION_FILE_REMOVE
)
def test_session_file_remove(
        desc, initial_content, session_name, exp_exc_type, exp_exc_msg):
    # pylint: disable=unused-argument
    """
    Test HMCSessionFile.remove().
    """
    session_file = create_session_file(initial_content)

    try:
        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                session_file.remove(session_name)

            exc = exc_info.value
            msg = str(exc)
            m = re.match(exp_exc_msg, msg)
            assert m, \
                "Unexpected exception message:\n" \
                "  expected pattern: {!r}\n" \
                "  actual message: {!r}".format(exp_exc_msg, msg)

            assert_session_file_content(session_file, initial_content)

        else:

            # The code to be tested
            session_file.remove(session_name)

            with pytest.raises(HMCSessionNotFound):
                session_file.get(session_name)

    finally:
        delete_session_file(session_file)


TESTCASES_SESSION_FILE_UPDATE = [

    # Each item is a testcase for testing HMCSessionFile.update(), with the
    # following items:
    #
    # * desc (str): Testcase description.
    # * initial_content (str): Content of initial HMC session file that is
    #   created before calling the function to be tested. None means there is
    #   no initial HMC session file.
    # * session_name (str): Session name to be removed.
    # * updates (dict): Dict with updates to the session.
    # * exp_exc_type (exc): Expected exception class. None for no exception.
    # * exp_exc_msg (str): Expected pattern for exception message. None for
    #   no exception.

    (
        "no session file",
        None,
        "foo",
        {},
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
    (
        "empty session file",
        "{}\n",
        "foo",
        {},
        HMCSessionNotFound,
        "Session not found in HMC session file"
    ),
    (
        "update nothing",
        """
foo:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "foo",
        {},  # creation_time gets updated automatically
        None, None
    ),
    (
        "update everything",
        """
foo:
  host: my_host
  userid: my_userid
  session_id: my_session_id
  ca_verify: false
  ca_cert_path: null
  creation_time: "2025-05-06 07:20:58"
        """,
        "foo",
        {
            "host": "my_host_2",
            "userid": "my_userid_2",
            "session_id": "my_session_id_2",
            "ca_verify": True,
            "ca_cert_path": "foo",
            # creation_time gets updated automatically
        },
        None, None
    ),
]


@pytest.mark.parametrize(
    "desc, initial_content, session_name, updates, exp_exc_type, exp_exc_msg",
    TESTCASES_SESSION_FILE_UPDATE
)
def test_session_file_update(
        desc, initial_content, session_name, updates, exp_exc_type,
        exp_exc_msg):
    # pylint: disable=unused-argument
    """
    Test HMCSessionFile.update().
    """
    session_file = create_session_file(initial_content)

    try:
        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                session_file.update(session_name, updates)

            exc = exc_info.value
            msg = str(exc)
            m = re.match(exp_exc_msg, msg)
            assert m, \
                "Unexpected exception message:\n" \
                "  expected pattern: {!r}\n" \
                "  actual message: {!r}".format(exp_exc_msg, msg)

            assert_session_file_content(session_file, initial_content)

        else:

            original_session = session_file.get(session_name)

            # The code to be tested
            session_file.update(session_name, updates)

            updated_session = session_file.get(session_name)
            exp_session = original_session
            exp_session.creation_time = now_str()

            def _update_exp_session(attr):
                if attr in updates:
                    setattr(exp_session, attr, updates[attr])

            _update_exp_session('host')
            _update_exp_session('userid')
            _update_exp_session('session_id')
            _update_exp_session('ca_verify')
            _update_exp_session('ca_cert_path')
            _update_exp_session('creation_time')
            assert_session_equal(updated_session, exp_session, session_name)

    finally:
        delete_session_file(session_file)



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for end2end tests.
"""


import os
import io
import json
import csv
import re
import subprocess
import random
import warnings
import tempfile
import pytest
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import
from zhmcclient.testutils import setup_hmc_session, teardown_hmc_session

from zhmccli._helper import CSV_DELIM, CSV_QUOTE, CSV_QUOTING_READ
from zhmccli._session_file import HMCSessionFile, HMCSession, \
    DEFAULT_SESSION_NAME


# Prefix used for names of resources that are created during tests
TEST_PREFIX = 'zhmcclient_tests_end2end'


def setup_session_file(hd):
    """
    Setup an HMC session file with a logged-on session to the HMC in the
    HMC definition.

    Parameters:
      hd (:class:`~zhmcclient.testutils.HMCDefinition`): The HMC definition of
        the HMC the test runs against.

    Returns:
      tuple(session, filepath): With:
      * session (zhmcclient.Session): Logged-on session.
      * filepath (str): Path name of HMC session file.
    """
    # Create an empty temporary file for use as the HMC session file
    home_dir = os.path.expanduser('~')
    # pylint: disable=consider-using-with
    file = tempfile.NamedTemporaryFile(
        mode='w', encoding="utf-8", delete=False,
        suffix='.yaml', prefix='.test_zhmc_', dir=home_dir)
    filepath = file.name
    file.write("{}")
    file.close()

    session_file = HMCSessionFile(filepath)

    # Make the zhmc commands use this session file
    os.environ['_ZHMC_TEST_SESSION_FILEPATH'] = filepath

    session = setup_hmc_session(hd)
    hmc_session = HMCSession.from_zhmcclient_session(session)
    session_file.add(DEFAULT_SESSION_NAME, hmc_session)

    return session, filepath


def teardown_session_file(session, filepath):
    """
    Tear down an HMC session file and its HMC session.

    Parameters:
      session (:class:`~zhmcclient.Session`): Logged-on session.
      filepath (str): Path name of HMC session file.
    """
    del os.environ['_ZHMC_TEST_SESSION_FILEPATH']
    teardown_hmc_session(session)
    os.remove(filepath)


@pytest.fixture(scope='module')
def zhmc_session(request, hmc_definition):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture that provides a logged-on zhmc session for an end2end test
    function.

    This function logs on to the HMC and creates an HMC session file with the
    logged on session.

    Using this fixture as an argument in a test function resolves to that
    session.

    Returns:
        zhmcclient.Session: Logged-on session.
    """
    session, filepath = setup_session_file(hmc_definition)

    yield session

    teardown_session_file(session, filepath)


def run_zhmc(args, pdb=False, log=False):
    """
    Run the zhmc command and return its exit code, stdout and stderr.

    Parameters:

      args (list/tuple of str): List of command line arguments, not including
        the 'zhmc' command name itself.

      pdb (bool): If True, debug the zhmc command. This is done by inserting
        the '--pdb' option to the command arguments, and by not capturing
        stdout/stderr.

      log (bool): If True and pdb is False, enable HMC logging for the zhmc
        command. This is done by inserting the '--log hmc=debug' option to the
        command arguments.

    Returns:
      tuple (rc, stdout, stderr) as follows:
      - rc (int): zhmc exit code
      - stdout (str): stdout string, or None if debugging the zhmc command
      - stderr (str): stderr string, or None if debugging the zhmc command
    """
    assert isinstance(args, (list, tuple))

    p_args = ['zhmc'] + args

    # Set up output capturing, dependent on pdb flag
    p_kwargs = {}
    if pdb:
        p_args.insert(1, '--pdb')
    else:
        p_kwargs['stdout'] = subprocess.PIPE
        p_kwargs['stderr'] = subprocess.PIPE
        if log:
            p_args.insert(1, 'all=debug')
            p_args.insert(1, '--log')

    # pylint: disable=consider-using-with
    proc = subprocess.Popen(p_args, **p_kwargs)

    stdout, stderr = proc.communicate()
    rc = proc.returncode
    if not pdb:
        stdout = stdout.decode()
        stderr = stderr.decode()

    return rc, stdout, stderr


def fixup(value, force_type=None):
    """
    Fix up CSV value written with csv.QUOTE_STRINGS and read with
    csv.QUOTE_MINIMAL.

    Parameters:
      value (str or ...): Value to be fixed up.
      force_type (type or None): Force the property to be converted to the
        specified type. If `None`, no forced type conversion happens.

    Returns:
      Fixed up value in correct type.
    """
    if force_type:
        if value is None and force_type is str:
            value = ''
        else:
            value = force_type(value)
    elif isinstance(value, str):
        if value == 'True':
            value = True
        elif value == 'False':
            value = False
        elif value == '':
            value = None  # Empty string can be created with force_type=str
        elif value[0] in ("[", "{"):
            # pylint: disable=eval-used
            value = eval(value)  # nosec
        else:
            try:
                value = int(value)
            except ValueError:
                try:
                    value = float(value)
                except ValueError:
                    pass
    return value


def parse_output(stdout, out_format, out_type):
    """
    Parse the standard output of the zhmc command into a Python list or dict.

    Parameters:

      stdout (str): Stdout string.

      out_format (str): Poutput format.
        Possible values: 'json', 'csv'.

      out_type (type): Expected Python type of the output after being parsed.
        Possible values: list, dict.

    Returns:
      type: Parsed python object
    """
    assert out_type in (list, dict)
    if out_format == 'json':
        out_obj = json.loads(stdout)
    else:
        assert out_format == 'csv'
        # Note that reading with csv.QUOTE_NONNUMERIC does not work, because
        # it attempts to convert unquoted boolean values to float. Any other
        # quoting value (e.g. csv.QUOTE_MINIMAL) returns any unquoted values
        # (strings, booleans, numerics) as string types.
        out_obj = list(csv.DictReader(
            io.StringIO(stdout),
            delimiter=CSV_DELIM, quotechar=CSV_QUOTE,
            quoting=CSV_QUOTING_READ))
        if out_type is dict:
            assert len(out_obj) == 1
            out_obj = out_obj[0]

        # Fix up datatypes created by csv.QUOTE_MINIMAL
        assert CSV_QUOTING_READ == csv.QUOTE_MINIMAL
        if out_type is list:
            for row in out_obj:
                for name in row:
                    row[name] = fixup(row[name])
        else:
            for name in out_obj:
                out_obj[name] = fixup(out_obj[name])

    assert isinstance(out_obj, out_type)
    return out_obj


def env2bool(name):
    """
    Evaluate the (string) value of the specified environment variable as a
    boolean value and return the result as a bool.

    The following variable values are considered True: 'true', 'yes', '1'.
    Any other value or if the variable is not set, is considered False.
    """
    return os.getenv(name, 'false').lower() in ('true', 'yes', '1')


class End2endTestWarning(UserWarning):
    """
    Python warning indicating an issue with an end2end test.
    """
    pass


def _res_name(item):
    """Return the resource name, used by pick_test_resources()"""
    if isinstance(item, (tuple, list)):
        return item[0].name
    return item.name


def pick_test_resources(res_list):
    """
    Return the list of resources to be tested.

    The env.var "TESTRESOURCES" controls which resources are picked for the
    test, as follows:

    * 'random': (default) one random choice from the input list of resources.
    * 'all': the complete input list of resources.
    * '<pattern>': The resources with names matching the regexp pattern.

    Parameters:
      res_list (list of zhmcclient.BaseResource or tuple thereof):
        List of resources to pick from. Tuple items are a resource and its
        parent resources.

    Returns:
      list of zhmcclient.BaseResource: Picked list of resources.
    """

    test_res = os.getenv('TESTRESOURCES', 'random')

    if test_res == 'random':
        return [random.choice(res_list)]  # nosec: B311

    if test_res == 'all':
        return sorted(res_list, key=_res_name)

    # match the pattern
    ret_list = []
    for item in res_list:
        if re.search(test_res, _res_name(item)):
            ret_list.append(item)
    return sorted(ret_list, key=_res_name)


def skipif_no_storage_mgmt_feature(cpc):
    """
    Skip the test if the "DPM Storage Management" feature is not enabled for
    the specified CPC, or if the CPC does not yet support it.
    """
    try:
        smf = cpc.feature_enabled('dpm-storage-management')
    except ValueError:
        smf = False
    if not smf:
        skip_warn("DPM Storage Mgmt feature not enabled or not supported "
                  "on CPC {c}".format(c=cpc.name))


def skipif_storage_mgmt_feature(cpc):
    """
    Skip the test if the "DPM Storage Management" feature is enabled for
    the specified CPC.
    """
    try:
        smf = cpc.feature_enabled('dpm-storage-management')
    except ValueError:
        smf = False
    if smf:
        skip_warn("DPM Storage Mgmt feature enabled on CPC {c}".
                  format(c=cpc.name))


def skip_warn(msg):
    """
    Issue an End2endTestWarning and skip the current pytest testcase with the
    specified message.
    """
    warnings.warn(msg, End2endTestWarning, stacklevel=2)
    pytest.skip(msg)



# Copyright 2016,2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HBAs.
"""


import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props
from ._cmd_partition import find_partition


def find_hba(cmd_ctx, client, cpc_name, partition_name, hba_name):
    """
    Find an HBA by name and return its resource object.
    """
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)
    try:
        hba = partition.hbas.find(name=hba_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return hba


@cli.group('hba', options_metavar=COMMAND_OPTIONS_METAVAR)
def hba_group():
    """
    Command group for managing HBAs (DPM mode only).

    The commands in this group work only on z13 CPCs in DPM mode.
    On z14 and later CPCs in DPM mode, HBAs are automatically managed.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@hba_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def hba_list(cmd_ctx, cpc, partition, **options):
    """
    List the HBAs in a partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_hba_list(cmd_ctx, cpc, partition, options))


@hba_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('HBA', type=str, metavar='HBA')
@click.pass_obj
def hba_show(cmd_ctx, cpc, partition, hba):
    """
    Show the details of an HBA.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_hba_show(cmd_ctx, cpc, partition, hba))


@hba_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.option('--name', type=str, required=True,
              help='The name of the new HBA.')
@click.option('--description', type=str, required=False,
              help='The description of the new HBA. '
              'Default: empty')
@click.option('--adapter', type=str, required=True,
              help='The name of the backing FCP adapter.')
@click.option('--port', type=str, required=True,
              help='The name of the port on the backing FCP adapter.')
@click.option('--device-number', type=str, required=False,
              help='The device number to be used for the new HBA. '
              'Default: auto-generated')
@click.pass_obj
def hba_create(cmd_ctx, cpc, partition, **options):
    """
    Create an HBA in a partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_hba_create(cmd_ctx, cpc, partition,
                                               options))


@hba_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('HBA', type=str, metavar='HBA')
@click.option('--name', type=str, required=False,
              help='The new name of the HBA.')
@click.option('--description', type=str, required=False,
              help='The new description of the HBA.')
@click.option('--device-number', type=str, required=False,
              help='The new device number to be used for the HBA.')
@click.pass_obj
def hba_update(cmd_ctx, cpc, partition, hba, **options):
    """
    Update the properties of an HBA.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_hba_update(cmd_ctx, cpc, partition, hba,
                                               options))


@hba_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('HBA', type=str, metavar='HBA')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the HBA.',
              prompt='Are you sure you want to delete this HBA ?')
@click.pass_obj
def hba_delete(cmd_ctx, cpc, partition, hba):
    """
    Delete an HBA.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_hba_delete(cmd_ctx, cpc, partition, hba))


def cmd_hba_list(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    if partition.hbas is None:
        hbas = []
    else:
        try:
            hbas = partition.hbas.list(filter_args=filter_args)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'name',
        'cpc',
        'partition',
    ]
    if not options['names_only']:
        show_list.extend([
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    cpc_additions = {}
    partition_additions = {}
    for hba in hbas:
        cpc_additions[hba.uri] = cpc_name
        partition_additions[hba.uri] = partition_name
    additions = {
        'cpc': cpc_additions,
        'partition': partition_additions,
    }

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, hbas, cmd_ctx.output_format, show_list,
                        additions, all=options['all'], sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_hba_show(cmd_ctx, cpc_name, partition_name, hba_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    hba = find_hba(cmd_ctx, client, cpc_name, partition_name, hba_name)

    try:
        hba.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(hba.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = partition_name

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_hba_create(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
        'port': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    adapter_name = org_options['adapter']
    try:
        adapter = partition.manager.cpc.adapters.find(name=adapter_name)
    except zhmcclient.NotFound:
        raise click_exception("Could not find adapter '{a}' in CPC '{c}'.".
                              format(a=adapter_name, c=cpc_name),
                              cmd_ctx.error_format)

    port_name = org_options['port']
    try:
        port = adapter.ports.find(name=port_name)
    except zhmcclient.NotFound:
        raise click_exception("Could not find port '{p}' on adapter '{a}' in "
                              "CPC '{c}'.".
                              format(p=port_name, a=adapter_name, c=cpc_name),
                              cmd_ctx.error_format)

    properties['adapter-port-uri'] = port.uri

    try:
        new_hba = partition.hbas.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New HBA '{h}' has been created.".
               format(h=new_hba.properties['name']))


def cmd_hba_update(cmd_ctx, cpc_name, partition_name, hba_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    hba = find_hba(cmd_ctx, client, cpc_name, partition_name, hba_name)

    org_options = original_options(options)
    properties = options_to_properties(org_options)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating HBA '{h}'.".
                   format(h=hba_name))
        return

    try:
        hba.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != hba_name:
        click.echo("HBA '{h}' has been renamed to '{hn}' and was updated.".
                   format(h=hba_name, hn=properties['name']))
    else:
        click.echo(f"HBA '{hba_name}' has been updated.")


def cmd_hba_delete(cmd_ctx, cpc_name, partition_name, hba_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    hba = find_hba(cmd_ctx, client, cpc_name, partition_name, hba_name)

    try:
        hba.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"HBA '{hba_name}' has been deleted.")



# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Support for an HMC session file in YAML format.

An HMC session file stores session-related data about logged-in HMC sessions.
"""

import os
import stat
from copy import deepcopy
from datetime import datetime
import yaml
import jsonschema

__all__ = ['HMCSessionFile', 'HMCSessionException',
           'HMCSessionNotFound', 'HMCSessionAlreadyExists',
           'HMCSessionFileNotFound', 'HMCSessionFileError',
           'HMCSessionFileFormatError', 'DEFAULT_SESSION_NAME']

# datetime format for creation time in HMC session file
TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

BLANKED_OUT_STRING = '********'

DEFAULT_SESSION_NAME = 'default'

HMC_SESSION_FILE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "JSON schema for an HMC session file in YAML format",
    "description": "List of logged-in HMC sessions",
    "type": "object",
    "additionalProperties": False,
    "patternProperties": {
        "^[a-z0-9_]+$": {
            "type": "object",
            "additionalProperties": False,
            "required": [
                "host",
                "userid",
                "session_id",
                "ca_verify",
                "ca_cert_path",
                "creation_time",
            ],
            "properties": {
                # Top-level group name:
                "host": {
                    "description": "HMC host, as hostname or IP address",
                    "type": "string",
                },
                "userid": {
                    "description": "HMC userid",
                    "type": "string",
                },
                "session_id": {
                    "description": "HMC session ID",
                    "type": "string",
                },
                "ca_verify": {
                    "description": "CA certificate validation is performed",
                    "type": "boolean",
                },
                "ca_cert_path": {
                    "description": "Path name of CA certificate file or "
                    "directory",
                    "type": ["string", "null"],
                },
                "creation_time": {
                    "description": "Creation time of the HMC session, in "
                    "'YYYY-MM-DD hh:mm:ss' format in the UTC timezone",
                    "type": "string",
                },
            },
        },
    },
}


class HMCSessionException(Exception):
    """
    Base class for errors with the HMC session file.
    """
    pass


class HMCSessionNotFound(HMCSessionException):
    """
    The HMC session with the specified name was not found in the HMC session
    file.
    """
    pass


class HMCSessionAlreadyExists(HMCSessionException):
    """
    The HMC session with the specified name already exists in the HMC session
    file.
    """
    pass


class HMCSessionFileNotFound(HMCSessionException):
    """
    The HMC session file was not found.
    """
    pass


class HMCSessionFileError(HMCSessionException):
    """
    Error reading or writing the HMC session file.
    """
    pass


class HMCSessionFileFormatError(HMCSessionException):
    """
    Error in the format of the content of the HMC session file.
    """
    pass


def now_str():
    """
    Return the current time as a string in 'YYYY-MM-DD hh:mm:ss' format
    in the UTC timezone, suitable for use in the HMC session file."
    """
    return datetime.utcnow().strftime(TIME_FORMAT)


class HMCSession:
    """
    Representation of an HMC session in the HMC session file.
    """

    def __init__(self, host, userid, session_id, ca_verify, ca_cert_path,
                 creation_time=None):
        """
        Parameters:
          host (str): HMC host, as hostname or IP address.
          userid (str): HMC userid.
          session_id (str): HMC session ID.
          ca_verify (bool): CA certificate validation is performed.
          ca_cert_path (str): Path name of CA certificate file or directory,
            or None if the default CA chain is used.
          creation_time (str): Creation time of the HMC session, in
            'YYYY-MM-DD hh:mm:ss' format in the UTC timezone. If `None`, it
            defaults to the current time.
        """
        self.host = host
        self.userid = userid
        self.session_id = session_id
        self.ca_verify = ca_verify
        self.ca_cert_path = ca_cert_path
        if creation_time is None:
            creation_time = now_str()
        self.creation_time = creation_time

    def __repr__(self):
        return (
            "HMCSession("
            f"host={self.host!r}, "
            f"userid={self.userid!r}, "
            f"session_id={BLANKED_OUT_STRING}, "
            f"ca_verify={self.ca_verify!r}, "
            f"ca_cert_path={self.ca_cert_path!r}, "
            f"creation_time={self.creation_time!r})")

    @staticmethod
    def from_zhmcclient_session(zhmcclient_session):
        """
        Return new HMCSession object from a zhmcclient Session.

        Parameters:
          zhmcclient_session (zhmcclient.Session): The zhmcclient session.

        Returns:
          HMCSession: new HMCSession object.
        """
        if zhmcclient_session.verify_cert is False:
            ca_verify = False
            ca_cert_path = None
        elif zhmcclient_session.verify_cert is True:
            ca_verify = True
            ca_cert_path = None
        else:
            ca_verify = True
            ca_cert_path = zhmcclient_session.verify_cert
        return HMCSession(
            zhmcclient_session.host,
            zhmcclient_session.userid,
            zhmcclient_session.session_id,
            ca_verify,
            ca_cert_path)

    def as_dict(self):
        """
        Return the HMC session properties as a dict.
        """
        return {
            "host": self.host,
            "userid": self.userid,
            "session_id": self.session_id,
            "ca_verify": self.ca_verify,
            "ca_cert_path": self.ca_cert_path,
            "creation_time": self.creation_time,
        }


def session_file_opener(path, flags):
    """
    Python opener function for the HMC session file.
    """
    return os.open(path, flags, mode=stat.S_IRUSR | stat.S_IWUSR)


class DictDot(dict):
    """Dict with dot representation."""

    def __repr__(self):
        return '{...}'


# The HMC session file that is normally used.
# Some unit/function tests specify another path name by passing the 'filepath'
# parameter to HMCSessionFile().
# Some end2end tests specify another path name by setting the
# _ZHMC_TEST_SESSION_FILEPATH env. var.
SESSION_FILEPATH = '~/.zhmc_sessions.yml'


class HMCSessionFile:
    """
    Access to an HMC session file.
    """

    def __init__(self, filepath=None):
        """
        Parameters:

          filepath (str): Path name of the HMC session file. If None, the
            path name in the env var _ZHMC_TEST_SESSION_FILEPATH is used. If
            not set, the path in SESSION_FILEPATH is used.
        """
        if filepath is None:
            filepath = os.environ.get('_ZHMC_TEST_SESSION_FILEPATH', None)
            if filepath is None:
                filepath = SESSION_FILEPATH
        self._filepath = os.path.expanduser(filepath)
        self._data = None  # File content, deferred loading

    def __repr__(self):
        if self._data is None:
            self._data = self._load()
        session_dict = {}
        for session_name in self._data.keys():
            session_dict[session_name] = DictDot()
        return (
            "HMCSessionFile("
            f"filepath={self.filepath!r}, "
            f"data={session_dict!r})")

    @property
    def filepath(self):
        """
        string: Path name of the HMC session file.
        """
        return self._filepath

    def list(self):
        """
        List all HMC sessions in the HMC session file.

        Returns:
          dict of session name, HMCSession

        Raises:
          HMCSessionFileError: HMC session file could not be created.
          HMCSessionFileFormatError: Invalid YAML syntax in HMC session file.
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        if self._data is None:
            self._data = self._load()
        sessions = {}
        for session_name, session_item in self._data.items():
            sessions[session_name] = HMCSession(**session_item)
        return sessions

    def get(self, name=DEFAULT_SESSION_NAME):
        """
        Get the HMC session with the specified name from the HMC session file.

        Parameters:
          name (str): Name of the HMC session.

        Returns:
          HMCSession: HMC session.

        Raises:
          HMCSessionNotFound: Session not found in HMC session file.
          HMCSessionFileError: HMC session file could not be created.
          HMCSessionFileFormatError: Invalid YAML syntax in HMC session file.
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        if self._data is None:
            self._data = self._load()
        try:
            session_item = self._data[name]
        except KeyError:
            raise HMCSessionNotFound(
                f"Session not found in HMC session file: {name}")
        return HMCSession(**session_item)

    def add(self, name, session):
        """
        Add the specified HMC session with the specified name to the HMC
        session file, and update the file.

        Parameters:
          name (str): Name of the HMC session.
          session (HMCSession): The HMC session to be added. The `creation_time`
            property is ignored if present, and is set to the current time in
            the HMC session file.

        Raises:
          HMCSessionAlreadyExists: HMC session already exists in session file.
          HMCSessionFileError: HMC session file could not be created.
          HMCSessionFileFormatError: Invalid YAML syntax in HMC session file.
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        if self._data is None:
            self._data = self._load()
        if name in self._data:
            raise HMCSessionAlreadyExists(
                f"Session already exists in HMC session file: {name}")
        self._data[name] = session.as_dict()
        self._data[name].update({'creation_time': now_str()})
        self._save(self._data)

    def remove(self, name):
        """
        Remove the specified HMC session from the HMC session file, and update
        the file.

        Parameters:
          name (str): Name of the HMC session.

        Raises:
          HMCSessionNotFound: Session not found in HMC session file.
          HMCSessionFileError: HMC session file could not be created.
          HMCSessionFileFormatError: Invalid YAML syntax in HMC session file.
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        if self._data is None:
            self._data = self._load()
        try:
            del self._data[name]
        except KeyError:
            raise HMCSessionNotFound(
                f"Session not found in HMC session file: {name}")
        self._save(self._data)

    def update(self, name, updates):
        """
        Update the specified HMC session with the provided keyword arguments,
        and update the file.

        Parameters:
          name (str): Name of the HMC session.
          updates (dict): Properties of the HMC session to be updated.
            The `creation_time` property is ignored if present, and is set to
            the current time in the HMC session file.

        Raises:
          HMCSessionNotFound: Session not found in HMC session file.
          HMCSessionFileError: HMC session file could not be created.
          HMCSessionFileFormatError: Invalid YAML syntax in HMC session file.
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        if self._data is None:
            self._data = self._load()
        data = deepcopy(self._data)
        try:
            session_data = data[name]
        except KeyError:
            raise HMCSessionNotFound(
                f"Session not found in HMC session file: {name}")
        session_data.update(updates)
        session_data.update({'creation_time': now_str()})
        self._save(data)
        self._data = data

    def _create(self):
        """
        Create an empty HMC session file and return its empty data.

        Raises:
          HMCSessionFileError: HMC session file could not be created.
        """
        try:
            with open(self._filepath, 'w', encoding="utf-8",
                      opener=session_file_opener) as fp:
                fp.write("{}")
            os.chmod(self._filepath, stat.S_IRUSR | stat.S_IWUSR)
        except OSError as exc:
            raise HMCSessionFileError(
                f"The HMC session file {self._filepath!r} could not be "
                f"created: {exc}")
        return {}

    def _load(self):
        """
        Load the HMC session file, validate it and return its data.

        Raises:
          HMCSessionFileError: HMC session file could not be created.
          HMCSessionFileFormatError: Invalid YAML syntax in HMC session file.
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        try:
            # pylint: disable=unspecified-encoding
            with open(self._filepath) as fp:
                try:
                    data = yaml.load(fp, Loader=yaml.SafeLoader)
                except (yaml.parser.ParserError,
                        yaml.scanner.ScannerError) as exc:
                    raise HMCSessionFileFormatError(
                        "Invalid YAML syntax in HMC session file "
                        f"{self._filepath!r}: {exc.__class__.__name__} {exc}")
        except OSError:
            data = self._create()
        self._validate(data)
        return data

    def _save(self, data):
        """
        Validate and save the data to the HMC session file.

        Raises:
          HMCSessionFileError: HMC session file could not be written.
        """
        self._validate(data)
        try:
            with open(self._filepath, 'w', encoding="utf-8",
                      opener=session_file_opener) as fp:
                yaml.dump(data, fp, indent=4, default_flow_style=False)
            os.chmod(self._filepath, stat.S_IRUSR | stat.S_IWUSR)
        except OSError as exc:
            raise HMCSessionFileError(
                f"The HMC session file {self._filepath!r} could not be "
                f"written: {exc}")

    def _validate(self, data):
        """
        Validate that the data conforms to the schema for the HMC session file.

        Raises:
          HMCSessionFileFormatError: Invalid data format in HMC session file.
        """
        try:
            jsonschema.validate(data, HMC_SESSION_FILE_SCHEMA)
        except jsonschema.exceptions.ValidationError as exc:
            elem = '.'.join(str(e) for e in exc.absolute_path)
            schemaitem = '.'.join(str(e) for e in exc.absolute_schema_path)
            raise HMCSessionFileFormatError(
                "Invalid data format in HMC session file "
                f"{self._filepath}: {exc.message}; "
                f"Offending element: {elem}; "
                f"Schema item: {schemaitem}; "
                f"Validator: {exc.validator}={exc.validator_value}")



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for global options.
"""


import urllib3
import pytest

from ..function.utils import call_zhmc_child, assert_rc, assert_patterns

urllib3.disable_warnings()

URLLIB3_VERSION = [int(v) for v in urllib3.__version__.split('.')]

if URLLIB3_VERSION < [2, 0]:
    INVALID_HOST_MSG = "Failed to establish a new connection:"
else:
    INVALID_HOST_MSG = "Failed to resolve"


@pytest.mark.parametrize(
    "err_format, exp_stderr_patterns", [
        (None,  # default format: msg
         [r"Error: ConnectionError: .*" + INVALID_HOST_MSG + r".*"]),
        ('msg',
         [r"Error: ConnectionError: .*" + INVALID_HOST_MSG + r".*"]),
        ('def',
         [r"Error: classname='ConnectionError'; message=['\"].*"
          + INVALID_HOST_MSG + r".*['\"];"]),  # noqa: W503
    ]
)
@pytest.mark.parametrize(
    "err_opt", ['-e', '--error-format']
)
def test_option_errorformat(err_opt, err_format, exp_stderr_patterns):
    """
    Test global option (-e, --error-format).
    """

    if err_format is None:
        err_args = []
    else:
        err_args = [err_opt, err_format]

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_child(
        err_args + ['info'],
        {'ZHMC_HOST': 'invalid_host', 'ZHMC_USERID': 'user'}
    )

    assert_rc(1, rc, stdout, stderr)
    assert stdout == ""
    assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Function tests for options of 'zhmc cpc dpm-import' command.
"""


from .utils import call_zhmc_child, assert_rc


def test_cpc_dpmimport_help():
    """Test 'zhmc cpc dpm-import --help'"""

    rc, stdout, stderr = call_zhmc_child(
        ['cpc', 'dpm-import', '--help'])

    assert_rc(0, rc, stdout, stderr)
    assert stdout.startswith(
        "Usage: zhmc cpc dpm-import [COMMAND-OPTIONS] CPC\n"), \
        f"stdout={stdout!r}"
    assert stderr == ""


def test_cpc_dpmimport_helpdpmfile():
    """Test 'zhmc cpc dpm-import --help-dpm-file'"""

    rc, stdout, stderr = call_zhmc_child(
        ['cpc', 'dpm-import', '--help-dpm-file'])

    assert_rc(0, rc, stdout, stderr)
    assert stdout.startswith(
        "\nFormat of DPM configuration file:\n"), \
        f"stdout={stdout!r}"
    assert stderr == ""


def test_cpc_dpmimport_helpmappingfile():
    """Test 'zhmc cpc dpm-import --help-mapping-file'"""

    rc, stdout, stderr = call_zhmc_child(
        ['cpc', 'dpm-import', '--help-mapping-file'])

    assert_rc(0, rc, stdout, stderr)
    assert stdout.startswith(
        "\nFormat of adapter mapping file:\n"), \
        f"stdout={stdout!r}"
    assert stderr == ""



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _helper module.
"""


import os
import re
from datetime import datetime, timezone, timedelta, date
import pytest
import click

from zhmcclient_mock import FakedSession

from zhmccli._helper import CmdContext, parse_yaml_flow_style, \
    parse_ec_levels, parse_adapter_names, parse_crypto_domains, \
    domains_to_domain_config, domain_config_to_props_list, \
    required_option, forbidden_option, required_envvar, bool_envvar, \
    parse_timestamp


# Test cases for parse_yaml_flow_style()
TESTCASES_PARSE_YAML_FLOW_STYLE = [
    # value, exp_obj, exp_exc_msg
    (
        '',
        None,
        None
    ),
    (
        'a',
        'a',
        None
    ),
    (
        '1',
        1,
        None
    ),
    (
        'null',
        None,
        None
    ),
    (
        '[]',
        [],
        None
    ),
    (
        '[a]',
        ['a'],
        None
    ),
    (
        '[a, b, 1, null]',
        ['a', 'b', 1, None],
        None
    ),
    (
        '{}',
        {},
        None
    ),
    (
        '{a: b}',
        {'a': 'b'},
        None
    ),
    (
        '{a: b, b: 1, c: null}',
        {'a': 'b', 'b': 1, 'c': None},
        None
    ),
    (
        '{',
        None,
        "Error parsing value of option '--option' in YAML FLow Collection Style"
    ),
]


@pytest.mark.parametrize(
    "value, exp_obj, exp_exc_msg",
    TESTCASES_PARSE_YAML_FLOW_STYLE)
def test_parse_yaml_flow_style(value, exp_obj, exp_exc_msg):
    """
    Test function for datetime_from_isoformat().
    """

    cmd_ctx = CmdContext(
        params=None, host='host', userid='host', password='password',
        no_verify=True, ca_certs=None, logon_source='options',
        session_name=None, output_format='table', transpose=False,
        error_format='msg', timestats=False, session_id=None, get_password=None,
        pdb=False)  # nosec: B106

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            parse_yaml_flow_style(cmd_ctx, '--option', value)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        obj = parse_yaml_flow_style(cmd_ctx, '--option', value)

        assert obj == exp_obj


# Test cases for test_parse_ec_levels()
TESTCASES_PARSE_EC_LEVELS = [
    # value, exp_obj, exp_exc_msg
    (
        '',
        None,
        "Error parsing value of option '--option': Value must be a list of "
        "strings: ''"
    ),
    (
        '[]',
        [],
        None
    ),
    (
        '[P12345.001]',
        [('P12345', '001')],
        None
    ),
    (
        '[P12345.001, P12346.002]',
        [('P12345', '001'), ('P12346', '002')],
        None
    ),
    (
        '[',
        None,
        "Error parsing value of option '--option' in YAML FLow Collection Style"
    ),
    (
        '{P12345: 001}',
        None,
        "Error parsing value of option '--option': Value must be a list of "
        "strings: '{P12345: 001}'"
    ),
    (
        'P12345.001',
        None,
        "Error parsing value of option '--option': Value must be a list of "
        "strings: 'P12345.001'"
    ),
    (
        '[P12345]',
        None,
        "Error parsing value of option '--option': Invalid EC level format "
        "'P12345'"
    ),
    (
        '[P12345.001.002]',
        None,
        "Error parsing value of option '--option': Invalid EC level format "
        "'P12345.001.002'"
    ),
]


@pytest.mark.parametrize(
    "value, exp_obj, exp_exc_msg",
    TESTCASES_PARSE_EC_LEVELS)
def test_parse_ec_levels(value, exp_obj, exp_exc_msg):
    """
    Test function for parse_ec_levels().
    """

    cmd_ctx = CmdContext(
        params=None, host='host', userid='host', password='password',
        no_verify=True, ca_certs=None, logon_source='options',
        session_name=None, output_format='table', transpose=False,
        error_format='msg', timestats=False, session_id=None, get_password=None,
        pdb=False)  # nosec: B106

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            parse_ec_levels(cmd_ctx, '--option', value)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        obj = parse_ec_levels(cmd_ctx, '--option', value)

        assert obj == exp_obj


# Test cases for test_parse_adapter_names()
TESTCASES_PARSE_ADAPTER_NAMES = [
    # value, exp_obj, exp_exc_msg
    (
        '',
        None,
        r"Error parsing value of option '--option': Value must be a list of "
        r"adapter names: "
    ),
    (
        '[]',
        [],
        None
    ),
    (
        '[HSM1]',
        ['HSM1'],
        None
    ),
    (
        '[HSM 1]',
        ['HSM 1'],
        None
    ),
    (
        '["HSM 1"]',
        ['HSM 1'],
        None
    ),
    (
        '[\'HSM 1\']',
        ['HSM 1'],
        None
    ),
    (
        '[HSM1, HSM2]',
        ['HSM1', 'HSM2'],
        None
    ),
    (
        '[',
        None,
        r"Error parsing value of option '--option' in YAML FLow Collection "
        r"Style"
    ),
    (
        '{HSM1: 1}',
        None,
        r"Error parsing value of option '--option': Value must be a list of "
        r"adapter names: '\{HSM1: 1\}'"
    ),
    (
        'HSM1',
        None,
        r"Error parsing value of option '--option': Value must be a list of "
        r"adapter names: 'HSM1'"
    ),
]


@pytest.mark.parametrize(
    "value, exp_obj, exp_exc_msg",
    TESTCASES_PARSE_ADAPTER_NAMES)
def test_parse_adapter_names(value, exp_obj, exp_exc_msg):
    """
    Test function for parse_adapter_names().
    """

    cmd_ctx = CmdContext(
        params=None, host='host', userid='host', password='password',
        no_verify=True, ca_certs=None, logon_source='options',
        session_name=None, output_format='table', transpose=False,
        error_format='msg', timestats=False, session_id=None, get_password=None,
        pdb=False)  # nosec: B106

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            parse_adapter_names(cmd_ctx, '--option', value)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        obj = parse_adapter_names(cmd_ctx, '--option', value)

        assert obj == exp_obj


# Test cases for test_parse_crypto_domains()
TESTCASES_PARSE_CRYPTO_DOMAINS = [
    # value, exp_obj, exp_exc_msg
    (
        '',
        None,
        r"Error parsing value of option '--option': Value must be a list of "
        r"domain index numbers or ranges thereof: "
    ),
    (
        '[]',
        [],
        None
    ),
    (
        '[12]',
        [12],
        None
    ),
    (
        '["12"]',
        [12],
        None
    ),
    (
        '[\'12\']',
        [12],
        None
    ),
    (
        '[xx]',
        None,
        r"Error parsing value of option '--option': Invalid integer value for "
        r"list item 'xx'"
    ),
    (
        '[0, 1]',
        [0, 1],
        None
    ),
    (
        '[12-14]',
        [12, 13, 14],
        None
    ),
    (
        '["12-14"]',
        [12, 13, 14],
        None
    ),
    (
        '[\'12-14\']',
        [12, 13, 14],
        None
    ),
    (
        '[12-12]',
        [12],
        None
    ),
    (
        '[11-12-13]',
        None,
        r"Error parsing value of option '--option': Invalid range format for "
        r"list item '11-12-13' - must be N-M"
    ),
    (
        '[12-11]',
        None,
        r"Error parsing value of option '--option': Invalid range values for "
        r"list item '12-11' - left value must not be larger than right value"
    ),
    (
        '[xx-12]',
        None,
        r"Error parsing value of option '--option': Invalid integer values for "
        r"range list item 'xx-12'"
    ),
    (
        '[[0]]',
        None,
        r"Error parsing value of option '--option': Invalid type for list item "
        r"\[0\] - must be string or int"
    ),
    (
        '[true]',
        None,
        r"Error parsing value of option '--option': Invalid type for list item "
        r"True - must be string or int"
    ),
    (
        '[false]',
        None,
        r"Error parsing value of option '--option': Invalid type for list item "
        r"False - must be string or int"
    ),
    (
        '[',
        None,
        r"Error parsing value of option '--option' in YAML FLow Collection "
        r"Style"
    ),
    (
        '{12: 13}',
        None,
        r"Error parsing value of option '--option': Value must be a list of "
        r"domain index numbers or ranges thereof: '\{12: 13\}'"
    ),
    (
        '12',
        None,
        r"Error parsing value of option '--option': Value must be a list of "
        r"domain index numbers or ranges thereof: '12'"
    ),
]


@pytest.mark.parametrize(
    "value, exp_obj, exp_exc_msg",
    TESTCASES_PARSE_CRYPTO_DOMAINS)
def test_parse_crypto_domains(value, exp_obj, exp_exc_msg):
    """
    Test function for parse_crypto_domains().
    """

    cmd_ctx = CmdContext(
        params=None, host='host', userid='host', password='password',
        no_verify=True, ca_certs=None, logon_source='options',
        session_name=None, output_format='table', transpose=False,
        error_format='msg', timestats=False, session_id=None, get_password=None,
        pdb=False)  # nosec: B106

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            parse_crypto_domains(cmd_ctx, '--option', value)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        obj = parse_crypto_domains(cmd_ctx, '--option', value)

        assert obj == exp_obj


# Test cases for test_domains_to_domain_config()
TESTCASES_DOMAINS_TO_DOMAIN_CONFIG = [
    # usage_domains, control_domains, exp_obj, exp_exc_msg
    (
        [],
        [],
        [],
        None,
    ),
    (
        [12],
        [],
        [
            {
                'domain-index': 12,
                'access-mode': 'control-usage',
            },
        ],
        None
    ),
    (
        [],
        [12],
        [
            {
                'domain-index': 12,
                'access-mode': 'control',
            },
        ],
        None
    ),
    (
        [12, 13],
        [14, 15],
        [
            {
                'domain-index': 12,
                'access-mode': 'control-usage',
            },
            {
                'domain-index': 13,
                'access-mode': 'control-usage',
            },
            {
                'domain-index': 14,
                'access-mode': 'control',
            },
            {
                'domain-index': 15,
                'access-mode': 'control',
            },
        ],
        None
    ),
]


@pytest.mark.parametrize(
    "usage_domains, control_domains, exp_obj, exp_exc_msg",
    TESTCASES_DOMAINS_TO_DOMAIN_CONFIG)
def test_domains_to_domain_config(
        usage_domains, control_domains, exp_obj, exp_exc_msg):
    """
    Test function for domains_to_domain_config().
    """

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            domains_to_domain_config(usage_domains, control_domains)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        obj = domains_to_domain_config(usage_domains, control_domains)

        assert obj == exp_obj


# Test cases for test_domain_config_to_props_list()
TESTCASES_DOMAIN_CONFIG_TO_PROPS_LIST = [
    # adapter_names, domain_configs, exp_props_list, exp_exc_msg
    (
        [],
        [],
        [],
        None,
    ),
    (
        ['A1'],
        [
            {
                'domain-index': 12,
                'access-mode': 'control-usage',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '12', 'access-mode': 'control-usage'},
        ],
        None
    ),
    (
        ['A1'],
        [
            {
                'domain-index': 0,
                'access-mode': 'control-usage',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '0', 'access-mode': 'control-usage'},
        ],
        None
    ),
    (
        ['A1'],
        [
            {
                'domain-index': 12,
                'access-mode': 'control',
            },
            {
                'domain-index': 13,
                'access-mode': 'control',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '12-13', 'access-mode': 'control'},
        ],
        None
    ),
    (
        ['A1'],
        [
            {
                'domain-index': 0,
                'access-mode': 'control',
            },
            {
                'domain-index': 1,
                'access-mode': 'control',
            },
            {
                'domain-index': 2,
                'access-mode': 'control',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '0-2', 'access-mode': 'control'},
        ],
        None
    ),
    (
        ['A1'],
        [
            {
                'domain-index': 12,
                'access-mode': 'control',
            },
            {
                'domain-index': 13,
                'access-mode': 'control',
            },
            {
                'domain-index': 14,
                'access-mode': 'control-usage',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '12-13', 'access-mode': 'control'},
            {'adapter': 'A1', 'domains': '14', 'access-mode': 'control-usage'},
        ],
        None
    ),
    (
        ['A1'],
        [
            {
                'domain-index': 12,
                'access-mode': 'control',
            },
            {
                'domain-index': 14,
                'access-mode': 'control',
            },
            {
                'domain-index': 15,
                'access-mode': 'control',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '12', 'access-mode': 'control'},
            {'adapter': 'A1', 'domains': '14-15', 'access-mode': 'control'},
        ],
        None
    ),
    (
        ['A1', 'A2'],
        [
            {
                'domain-index': 14,
                'access-mode': 'control',
            },
            {
                'domain-index': 15,
                'access-mode': 'control',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '14-15', 'access-mode': 'control'},
            {'adapter': 'A2', 'domains': '14-15', 'access-mode': 'control'},
        ],
        None
    ),
    (
        ['A2', 'A1'],
        [
            {
                'domain-index': 15,
                'access-mode': 'control',
            },
            {
                'domain-index': 14,
                'access-mode': 'control',
            },
        ],
        [
            {'adapter': 'A1', 'domains': '14-15', 'access-mode': 'control'},
            {'adapter': 'A2', 'domains': '14-15', 'access-mode': 'control'},
        ],
        None
    ),
]


@pytest.mark.parametrize(
    "adapter_names, domain_configs, exp_props_list, exp_exc_msg",
    TESTCASES_DOMAIN_CONFIG_TO_PROPS_LIST)
def test_domain_config_to_props_list(
        adapter_names, domain_configs, exp_props_list, exp_exc_msg):
    """
    Test function for domain_config_to_props_list().
    """

    session = FakedSession('fake-host', 'fake-hmc', '2.16.0', '4.10')
    cpc = session.hmc.cpcs.add({'name': 'cpc1'})

    adapters = []
    for adapter_name in adapter_names:
        adapter = cpc.adapters.add({'name': adapter_name, 'type': 'crypto'})
        adapters.append(adapter)

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            domain_config_to_props_list(adapters, 'adapter', domain_configs)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        props_list = domain_config_to_props_list(
            adapters, 'adapter', domain_configs)

        assert props_list == exp_props_list


# Test cases for required_option()
TESTCASES_REQUIRED_OPTION = [
    # value, name, exp_value, exp_exc_msg
    (
        'v1',
        'o1',
        'v1',
        None
    ),
    (
        '',
        'o1',
        '',
        None
    ),
    (
        None,
        'o1',
        None,
        "Required option not specified: o1"
    ),
]


@pytest.mark.parametrize(
    "value, name, exp_value, exp_exc_msg",
    TESTCASES_REQUIRED_OPTION)
def test_required_option(value, name, exp_value, exp_exc_msg):
    """
    Test function for required_option().
    """

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            required_option(value, name)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        act_value = required_option(value, name)

        assert act_value == exp_value


# Test cases for forbidden_option()
TESTCASES_FORBIDDEN_OPTION = [
    # value, name, reason, exp_exc_msg
    (
        None,
        'o1',
        'of some reason',
        None
    ),
    (
        '',
        'o1',
        'of some reason',
        "Option is not allowed because of some reason: o1"
    ),
    (
        'v1',
        'o1',
        'of some reason',
        "Option is not allowed because of some reason: o1"
    ),
]


@pytest.mark.parametrize(
    "value, name, reason, exp_exc_msg",
    TESTCASES_FORBIDDEN_OPTION)
def test_forbidden_option(value, name, reason, exp_exc_msg):
    """
    Test function for forbidden_option().
    """

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            forbidden_option(value, name, reason)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        forbidden_option(value, name, reason)


# Test cases for required_envvar()
TESTCASES_REQUIRED_ENVVAR = [
    # initial_value, name, exp_exc_msg
    (
        None,
        'e1',
        "Required environment variable not set: e1"
    ),
    (
        '',
        'e1',
        None
    ),
    (
        'v1',
        'e1',
        None
    ),
]


@pytest.mark.parametrize(
    "initial_value, name, exp_exc_msg",
    TESTCASES_REQUIRED_ENVVAR)
def test_required_envvar(initial_value, name, exp_exc_msg):
    """
    Test function for required_envvar().
    """

    if initial_value is not None:
        os.environ[name] = initial_value
    else:
        if name in os.environ:
            del os.environ[name]

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            required_envvar(name)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        value = required_envvar(name)

        assert value == initial_value


# Test cases for bool_envvar()
TESTCASES_BOOL_ENVVAR = [
    # initial_value, name, default, exp_value, exp_exc_msg
    (
        None,
        'e1',
        'd1',
        'd1',
        None,
    ),
    (
        None,
        'e1',
        None,
        None,
        None,
    ),
    (
        '0',
        'e1',
        'd1',
        False,
        None
    ),
    (
        'no',
        'e1',
        'd1',
        False,
        None
    ),
    (
        'No',
        'e1',
        'd1',
        False,
        None
    ),
    (
        'NO',
        'e1',
        'd1',
        False,
        None
    ),
    (
        'false',
        'e1',
        'd1',
        False,
        None
    ),
    (
        'False',
        'e1',
        'd1',
        False,
        None
    ),
    (
        'FALSE',
        'e1',
        'd1',
        False,
        None
    ),
    (
        '1',
        'e1',
        'd1',
        True,
        None
    ),
    (
        'yes',
        'e1',
        'd1',
        True,
        None
    ),
    (
        'Yes',
        'e1',
        'd1',
        True,
        None
    ),
    (
        'YES',
        'e1',
        'd1',
        True,
        None
    ),
    (
        'true',
        'e1',
        'd1',
        True,
        None
    ),
    (
        'True',
        'e1',
        'd1',
        True,
        None
    ),
    (
        'TRUE',
        'e1',
        'd1',
        True,
        None
    ),
    (
        '',
        'e1',
        'd1',
        None,
        "Invalid value for e1 environment variable: '' is not a valid boolean."
    ),
    (
        '2',
        'e1',
        'd1',
        None,
        "Invalid value for e1 environment variable: '2' is not a "
        "valid boolean."
    ),
    (
        'yesssir',
        'e1',
        'd1',
        None,
        "Invalid value for e1 environment variable: 'yesssir' is not a "
        "valid boolean."
    ),
]


@pytest.mark.parametrize(
    "initial_value, name, default, exp_value, exp_exc_msg",
    TESTCASES_BOOL_ENVVAR)
def test_bool_envvar(initial_value, name, default, exp_value, exp_exc_msg):
    """
    Test function for bool_envvar().
    """

    if initial_value is not None:
        os.environ[name] = initial_value
    else:
        if name in os.environ:
            del os.environ[name]

    if exp_exc_msg:
        with pytest.raises(click.exceptions.ClickException) as exc_info:

            # The function to be tested
            bool_envvar(name, default)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        value = bool_envvar(name, default)

        assert value == exp_value


TESTCASES_PARSE_TIMESTAMP = [
    # Test cases for parse_timestamp(). Each item is a testcase with the
    # following tuple items:
    # * value (str): Timestamp value to be parsed
    # * exp_dt (datetime): Expected parsed datetime value, if success.
    # * exp_exc_type (Exception): Expected excaption class, if error.
    # * exp_exc_msg (str): Expected exception message pattern, if error.
    (
        None,
        None,
        TypeError,
        ".*"
    ),
    (
        1,
        None,
        TypeError,
        ".*"
    ),
    (
        '',
        None,
        ValueError,
        "String does not contain a date"
    ),
    (
        '-1',
        None,
        ValueError,
        "Negative HMC timestamp value .* cannot be represented"
    ),
    (
        '2025-04-20 08:13:38+-!0000',
        None,
        ValueError,
        "Unknown string format"
    ),
    (
        '2025000-04-20',
        None,
        ValueError,
        "year .* is out of range"
    ),
    (
        '2025000000000000-04-20',
        None,
        OverflowError,
        "Python int too large|signed integer is greater than maximum"
    ),
    (
        '2025000000000000000000000000000000-04-20',
        None,
        OverflowError,
        "Python int too large|signed integer is greater than maximum"
    ),
    (
        '0',
        datetime(1970, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '1',
        datetime(1970, 1, 1, 0, 0, 0, 1000, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '1745136818344',
        datetime(2025, 4, 20, 8, 13, 38, 344000, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20 08:13:38.344+0000',
        datetime(2025, 4, 20, 8, 13, 38, 344000, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20 08:13:38.344+0200',
        datetime(2025, 4, 20, 8, 13, 38, 344000,
                 tzinfo=timezone(timedelta(seconds=7200))),
        None,
        None
    ),
    (
        '2025-04-20 08:13:38.344',
        datetime(2025, 4, 20, 8, 13, 38, 344000, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20 08:13:38',
        datetime(2025, 4, 20, 8, 13, 38, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20 08:13',
        datetime(2025, 4, 20, 8, 13, 0, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20 08',
        datetime(2025, 4, 20, 8, 0, 0, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20',
        datetime(2025, 4, 20, 0, 0, 0, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '05-04-06',
        datetime(2006, 5, 4, 0, 0, 0, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '5/4/06',
        datetime(2006, 5, 4, 0, 0, 0, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '08:13:38',
        datetime(date.today().year, date.today().month, date.today().day,
                 8, 13, 38, 0, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20T08:13:38.344Z',
        datetime(2025, 4, 20, 8, 13, 38, 344000, tzinfo=timezone.utc),
        None,
        None
    ),
    (
        '2025-04-20T08:13:38.344+02:00',
        datetime(2025, 4, 20, 8, 13, 38, 344000,
                 tzinfo=timezone(timedelta(seconds=7200))),
        None,
        None
    ),
    (
        '2025-04-20T08:13:38.344',
        datetime(2025, 4, 20, 8, 13, 38, 344000, tzinfo=timezone.utc),
        None,
        None
    ),
]


@pytest.mark.parametrize(
    "value, exp_dt, exp_exc_type, exp_exc_msg",
    TESTCASES_PARSE_TIMESTAMP)
def test_parse_timestamp(value, exp_dt, exp_exc_type, exp_exc_msg):
    """
    Test function for parse_timestamp().
    """

    if exp_exc_type:
        with pytest.raises(exp_exc_type) as exc_info:

            # The function to be tested
            parse_timestamp(value)

        exc = exc_info.value
        msg = str(exc)
        m = re.match(exp_exc_msg, msg)
        assert m, \
            "Unexpected exception message:\n" \
            "  expected pattern: {!r}\n" \
            "  actual message: {!r}".format(exp_exc_msg, msg)
    else:

        # The function to be tested
        dt = parse_timestamp(value)

        assert dt == exp_dt



# Copyright 2017,2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Function tests for 'zhmc info' command, using faked HMCs (zhmcclient mock).
"""

from .utils import call_zhmc_child, assert_rc


def test_info_help():
    """
    Test 'zhmc info --help'
    """

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_child(['info', '--help'])

    assert_rc(0, rc, stdout, stderr)
    assert stdout.startswith(
        "Usage: zhmc info [OPTIONS]\n"), \
        f"stdout={stdout!r}"
    assert stderr == ""



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for LDAP server definition management.
"""


import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props


def find_ldapdef(cmd_ctx, console, ldapdef_name):
    """
    Find a LDAP server definition by name and return its resource object.
    """
    try:
        ldapdef = console.ldap_server_definitions.find(
            name=ldapdef_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return ldapdef


@cli.group('ldap', options_metavar=COMMAND_OPTIONS_METAVAR)
def ldapdef_group():
    """
    Command group for managing LDAP server definitions.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@ldapdef_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def ldapdef_list(cmd_ctx, **options):
    """
    List the LDAP server definitions.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_ldapdef_list(cmd_ctx, options))


@ldapdef_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('LDAPSD', type=str, metavar='LDAP_SERVER_DEFINITION')
@click.pass_obj
def ldapdef_show(cmd_ctx, ldapsd):
    """
    Show the details of an LDAP server definition.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Console.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_ldapdef_show(cmd_ctx, ldapsd))


@ldapdef_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, required=True,
              help='The name of the new LDAP server definition.')
@click.option('--description', type=str, required=False,
              help='The description of the new LDAP server definition.')
@click.option('--primary-server', type=str, required=True,
              help='The host name or IP address of the primary LDAP server.')
@click.option('--port', type=int, required=False,
              help='The TCP port number of the LDAP servers. Specify 0 to set '
              'to the default of null, which causes the HMC to use the '
              'standard LDAP ports 636 for SSL and 389 for non-SSL.')
@click.option('--backup-server', type=str, required=False,
              help='The host name or IP address of the backup LDAP '
              'server. Specify the empty string to unset the backup '
              'LDAP server. Default: No backup LDAP server.')
@click.option('--use-ssl', type=bool, required=False,
              help='The indicator whether the LDAP servers use SSL. '
              'Default: False')
@click.option('--tolerate-untrusted-certificates', type=bool, required=False,
              help='The indicator whether the LDAP servers should '
              'tolerate self-signed or otherwise untrusted certificates. '
              'Default: False')
@click.option('--bind-distinguished-name', type=str, required=False,
              help='The distinguished name to bind to on the LDAP servers, '
              'or the empty string for anonymous bind. '
              'Default: Anonymous bind.')
@click.option('--bind-password', type=str, required=False,
              help='The password to be used when binding to a '
              'distinguished name on the LDAP servers. Specify the empty '
              'string for anonymous bind. '
              'Default: Anonymous bind.')
@click.option('--location-method', type=click.Choice(['pattern', 'subtree']),
              required=False,
              help='The method the LDAP servers use to locate a user\'s '
              'directory entry. '
              'Default: "pattern"')
@click.option('--search-distinguished-name', type=str, required=True,
              help='The distinguished name to use when searching for a '
              'user\'s directory entry. Use depends on location method.')
@click.option('--search-scope', type=click.Choice(['all', 'one-level']),
              required=False,
              help='The indicator how much of the subtree should be '
              'searched when searching for a user\'s directory entry in a '
              'subtree. Only for location method "subtree". '
              'Default: "all".')
@click.option('--search-filter', type=str, required=False,
              help='The LDAP search filter to use when searching for a '
              'user\'s directory entry in a subtree. Required for location '
              'method "subtree".')
@click.pass_obj
def ldapdef_create(cmd_ctx, **options):
    """
    Create a user-defined LDAP server definition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_ldapdef_create(cmd_ctx, options))


@ldapdef_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('LDAPSD', type=str, metavar='LDAP_SERVER_DEFINITION')
@click.option('--description', type=str, required=False,
              help='The new description of the LDAP server definition.')
@click.option('--primary-server', type=str, required=False,
              help='The new host name or IP address of the primary LDAP '
              'server.')
@click.option('--port', type=int, required=False,
              help='The new TCP port number of the LDAP servers. Specify 0 to '
              'set to the default of null, which causes the HMC to use the '
              'standard LDAP ports 636 for SSL and 389 for non-SSL.')
@click.option('--backup-server', type=str, required=False,
              help='The new host name or IP address of the backup LDAP '
              'server. Specify the empty string to unset the backup LDAP '
              'server.')
@click.option('--use-ssl', type=bool, required=False,
              help='The new indicator whether the LDAP server uses SSL.')
@click.option('--tolerate-untrusted-certificates', type=bool, required=False,
              help='The new indicator whether the LDAP servers should '
              'tolerate self-signed or otherwise untrusted certificates.')
@click.option('--bind-distinguished-name', type=str, required=False,
              help='The new distinguished name to bind to on the LDAP servers, '
              'or the empty string for anonymous bind.')
@click.option('--bind-password', type=str, required=False,
              help='The new password to be used when binding to a '
              'distinguished name on the LDAP servers. Specify the empty '
              'string for anonymous bind.')
@click.option('--location-method', type=click.Choice(['pattern', 'subtree']),
              required=False,
              help='The new method the LDAP servers use to locate a user\'s '
              'directory entry.')
@click.option('--search-distinguished-name', type=str, required=False,
              help='The new distinguished name to use when searching for a '
              'user\'s directory entry. Use depends on location method.')
@click.option('--search-scope', type=click.Choice(['all', 'one-level']),
              required=False,
              help='The new indicator how much of the subtree should be '
              'searched when searching for a user\'s directory entry in a '
              'subtree. Only for location method "subtree".')
@click.option('--search-filter', type=str, required=False,
              help='The new LDAP search filter to use when searching for a '
              'user\'s directory entry in a subtree. Only for location method '
              '"subtree".')
@click.pass_obj
def ldapdef_update(cmd_ctx, ldapsd, **options):
    """
    Update the properties of an LDAP server definition.

    Note that LDAP server definitions cannot be renamed.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_ldapdef_update(cmd_ctx, ldapsd, options))


@ldapdef_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('LDAPSD', type=str, metavar='LDAP_SERVER_DEFINITION')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the LDAP server '
              'definition.',
              prompt='Are you sure you want to delete this LDAP server '
              'definition ?')
@click.pass_obj
def ldapdef_delete(cmd_ctx, ldapsd):
    """
    Delete a user-defined LDAP server definition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_ldapdef_delete(cmd_ctx, ldapsd))


def cmd_ldapdef_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'description',
            'location-method',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    additions = {}

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    try:
        ldapdefs = console.ldap_server_definitions.list(
            full_properties=False, filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, ldapdefs, cmd_ctx.output_format,
                        show_list, additions, all=options['all'],
                        sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_ldapdef_show(cmd_ctx, ldapdef_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    ldapdef = find_ldapdef(cmd_ctx, console, ldapdef_name)

    try:
        ldapdef.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(ldapdef.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = console.name

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_ldapdef_create(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    name_map = {
        'primary-server': 'primary-hostname-ipaddr',
        'backup-server': 'backup-hostname-ipaddr',
        'port': 'connection-port',
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    # Handle special values
    for prop in ('connection-port',):
        if prop in properties and properties[prop] == 0:
            properties[prop] = None
    for prop in ('backup-hostname-ipaddr', 'bind-distinguished-name',
                 'bind-password'):
        if prop in properties and properties[prop] == '':
            properties[prop] = None

    try:
        new_ldapdef = console.ldap_server_definitions.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New LDAP server definition '{lsd}' has been created.".
               format(lsd=new_ldapdef.properties['name']))


def cmd_ldapdef_update(cmd_ctx, ldapdef_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    ldapdef = find_ldapdef(cmd_ctx, console, ldapdef_name)

    name_map = {
        'primary-server': 'primary-hostname-ipaddr',
        'backup-server': 'backup-hostname-ipaddr',
        'port': 'connection-port',
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    # Handle special values
    for prop in ('connection-port',):
        if prop in properties and properties[prop] == 0:
            properties[prop] = None
    for prop in ('backup-hostname-ipaddr', 'bind-distinguished-name',
                 'bind-password'):
        if prop in properties and properties[prop] == '':
            properties[prop] = None

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating LDAP server "
                   "definition '{lsd}'.".
                   format(lsd=ldapdef_name))
        return

    try:
        ldapdef.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("LDAP server definition '{lsd}' has been updated.".
               format(lsd=ldapdef_name))


def cmd_ldapdef_delete(cmd_ctx, ldapdef_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    ldapdef = find_ldapdef(cmd_ctx, console, ldapdef_name)

    try:
        ldapdef.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("LDAP server definition '{lsd}' has been deleted.".
               format(lsd=ldapdef_name))



# Copyright 2016,2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main script.
"""


import os
import sys
import logging
from logging.handlers import SysLogHandler
from logging import StreamHandler, NullHandler
import platform
import urllib3
import click
import click_repl
from prompt_toolkit.history import FileHistory

import zhmcclient
import zhmcclient_mock
from ._session_file import HMCSessionFile, HMCSessionNotFound, \
    DEFAULT_SESSION_NAME


from ._helper import CmdContext, GENERAL_OPTIONS_METAVAR, REPL_HISTORY_FILE, \
    REPL_PROMPT, OUTPUT_FORMATS, TABLE_FORMATS, LOG_LEVELS, LOG_DESTINATIONS, \
    SYSLOG_FACILITIES, click_exception, get_click_terminal_width, \
    required_option, forbidden_option, required_envvar, bool_envvar, \
    LogonSource


urllib3.disable_warnings()


# Default values for some options
DEFAULT_OUTPUT_FORMAT = 'table'
DEFAULT_ERROR_FORMAT = 'msg'
DEFAULT_TIMESTATS = False
DEFAULT_LOG = 'all=warning'
DEFAULT_LOG_DESTINATION = 'stderr'
DEFAULT_SYSLOG_FACILITY = 'user'
DEFAULT_NO_VERIFY = False

CONSOLE_LOGGER_NAME = 'zhmccli.console'

ERROR_FORMATS = ['msg', 'def']

# List of values to try for the 'address' parameter when creating
# a SysLogHandler object.
# Key: Operating system type, as returned by platform.system(). For CygWin,
# the returned value is 'CYGWIN_NT-6.1', which is special-cased to 'CYGWIN_NT'.
# Value: List of values for the 'address' parameter; to be tried in the
# specified order.
SYSLOG_ADDRESSES = {
    'Linux': ['/dev/log', ('localhost', 514)],
    'Darwin': ['/var/run/syslog', ('localhost', 514)],  # OS-X
    'Windows': [('localhost', 514)],
    'CYGWIN_NT': ['/dev/log', ('localhost', 514)],  # Requires syslog-ng pkg
}

# The getattr() is used to work around a Pylint false positive no-member issue
ZHMCCLIENT_VERSION = "zhmcclient, version {v}".format(
    v=getattr(zhmcclient, '__version__', 'unknown'))

# Logger names by log component
LOGGER_NAMES = {
    'all': '',  # root logger
    'api': zhmcclient.API_LOGGER_NAME,
    'hmc': zhmcclient.HMC_LOGGER_NAME,
    'console': CONSOLE_LOGGER_NAME,
}
LOG_COMPONENTS = LOGGER_NAMES.keys()

# Context variables passed to Click
CLICK_CONTEXT_SETTINGS = dict(

    # Set the terminal width - used e.g. for Click help messages
    terminal_width=get_click_terminal_width(),
)


@click.group(invoke_without_command=True,
             context_settings=CLICK_CONTEXT_SETTINGS,
             options_metavar=GENERAL_OPTIONS_METAVAR)
@click.option('-h', '--host', type=str, default=None,
              help="Hostname or IP address of the HMC. "
                   "This option determines where the logon data (except "
                   "password) comes from: "
                   "If specified, logon data comes from command line options. "
                   "Otherwise if the ZHMC_HOST environment variable is set, "
                   "logon data comes from ZHMC_* environment variables. "
                   "Otherwise, logon data comes from the HMC session file.")
@click.option('-u', '--userid', type=str, default=None,
              help="Username for the HMC. "
                   "The corresponding environment variable is ZHMC_USERID.")
@click.option('-p', '--password', type=str, envvar='ZHMC_PASSWORD',
              help="Password for the HMC. If not specified, it is taken from "
                   "the ZHMC_PASSWORD environment variable. If that is also "
                   "not set, it is prompted for. Note that this approach for "
                   "the password is independent of where the other logon data "
                   "comes from.")
@click.option('-n', '--no-verify', is_flag=True, default=None,
              help="Do not verify the HMC certificate. "
                   "The corresponding environment variable is ZHMC_NO_VERIFY, "
                   "with values 0/no/false or 1/yes/true.")
@click.option('-c', '--ca-certs', type=str, default=None,
              help="Path name of certificate file or directory with CA "
                   "certificates to be used for verifying the HMC certificate. "
                   "The corresponding environment variable is ZHMC_CA_CERTS. "
                   "The empty string causes the path name to be taken from "
                   "the REQUESTS_CA_BUNDLE environment variable, "
                   "or from the CURL_CA_BUNDLE environment variable, "
                   "or the 'certifi' Python package is used which provides the "
                   "Mozilla CA Certificate List).")
@click.option('-s', '--session-name', type=str, metavar='NAME', default=None,
              help="Session name when using the HMC session file. "
              f"(Default: {DEFAULT_SESSION_NAME})")
@click.option('-o', '--output-format',
              type=click.Choice(OUTPUT_FORMATS),
              help='Output format (Default: {def_of}).'.
              format(def_of=DEFAULT_OUTPUT_FORMAT))
@click.option('-x', '--transpose', type=str, is_flag=True,
              help='Transpose the output table for metrics.')
@click.option('-e', '--error-format', type=click.Choice(ERROR_FORMATS),
              help='Error message format (Default: {def_ef}).'.
              format(def_ef=DEFAULT_ERROR_FORMAT))
@click.option('-t', '--timestats', type=str, is_flag=True,
              help='Show time statistics of HMC operations.')
@click.option('--log', type=str, metavar='COMP=LEVEL,...',
              help="Set a component to a log level (COMP: [{comps}], "
              "LEVEL: [{levels}], Default: {def_log}).".
              format(comps='|'.join(LOG_COMPONENTS),
                     levels='|'.join(LOG_LEVELS),
                     def_log=DEFAULT_LOG))
@click.option('--log-dest', type=str,
              metavar='[{}]'.format('|'.join(LOG_DESTINATIONS)),
              help="Log destination for this command (Default: {def_dest}).".
              format(def_dest=DEFAULT_LOG_DESTINATION))
@click.option('--syslog-facility', type=click.Choice(SYSLOG_FACILITIES),
              help="Syslog facility when logging to the syslog "
              "(Default: {def_slf}).".
              format(def_slf=DEFAULT_SYSLOG_FACILITY))
@click.option('--pdb', is_flag=True, hidden=True,
              help='Break execution in the pdb debugger just before '
                   'executing the command within zhmc.')
@click.version_option(
    message='%(prog)s, version %(version)s\n' + ZHMCCLIENT_VERSION,
    help="Show the versions of this command and of the zhmcclient package and "
    "exit.")
@click.pass_context
def cli(ctx, host, userid, password, no_verify, ca_certs, session_name,
        output_format, transpose, error_format, timestats, log, log_dest,
        syslog_facility, pdb):
    """
    Command line interface for the IBM Z HMC.

    The options shown in this help text are general options that can also
    be specified on any of the (sub-)commands.
    """

    # Concept: In interactive mode, the global options specified in the command
    # line are used as defaults for the commands that are issued interactively.
    # The interactive commands may override these options.
    # This requires being able to determine for each option whether it has been
    # specified. This is the reason the options don't define defaults in the
    # decorators that define them.

    if ctx.obj is None:
        # We are in command mode or are processing the command line options in
        # interactive mode.
        # We apply the documented option defaults.
        if session_name is None:
            session_name = DEFAULT_SESSION_NAME
        if host is not None:
            # Logon data comes from command line options
            logon_source = LogonSource.OPTIONS
            userid = required_option(userid, '--userid')
            if no_verify is None:
                no_verify = DEFAULT_NO_VERIFY
            session_id = None
        elif os.getenv('ZHMC_HOST') is not None:
            # Logon data comes from command ZHMC_* environment variables
            logon_source = LogonSource.ENVIRONMENT
            reason = "logon data from ZHMC_* environment variables is used"
            forbidden_option(userid, '--userid', reason)
            forbidden_option(no_verify, '--no-verify', reason)
            forbidden_option(ca_certs, '--ca-certs', reason)
            host = required_envvar('ZHMC_HOST')
            userid = required_envvar('ZHMC_USERID')
            no_verify = bool_envvar('ZHMC_NO_VERIFY', DEFAULT_NO_VERIFY)
            ca_certs = os.getenv('ZHMC_CA_CERTS', None)
            session_id = os.getenv('ZHMC_SESSION_ID', None)
        else:
            # Logon data comes from HMC session file
            reason = "logon data from HMC session file is used"
            forbidden_option(userid, '--userid', reason)
            forbidden_option(no_verify, '--no-verify', reason)
            forbidden_option(ca_certs, '--ca-certs', reason)
            session_file = HMCSessionFile()
            try:
                hmc_session = session_file.get(session_name)
            except HMCSessionNotFound:
                logon_source = LogonSource.NONE
                host = None
                userid = None
                no_verify = None
                ca_certs = None
                session_id = None
            else:
                logon_source = LogonSource.SESSION_FILE
                host = hmc_session.host
                userid = hmc_session.userid
                no_verify = not hmc_session.ca_verify
                ca_certs = hmc_session.ca_cert_path
                session_id = hmc_session.session_id

        if output_format is None:
            output_format = DEFAULT_OUTPUT_FORMAT
        if transpose is None:
            transpose = False
        if error_format is None:
            error_format = DEFAULT_ERROR_FORMAT
        if timestats is None:
            timestats = DEFAULT_TIMESTATS
    else:
        # We are processing an interactive command.
        # We apply the option defaults from the command line options.
        logon_source = ctx.obj.logon_source
        session_id = ctx.obj.session_id
        if host is None:
            host = ctx.obj.host
        if userid is None:
            userid = ctx.obj.userid
        if no_verify is None:
            no_verify = ctx.obj.no_verify
        if ca_certs is None:
            ca_certs = ctx.obj.ca_certs
        if password is None:
            # pylint: disable=protected-access
            password = ctx.obj._password
        if session_name is None:
            session_name = ctx.obj.session_name
        if output_format is None:
            output_format = ctx.obj.output_format
        if transpose is None:
            transpose = ctx.obj.transpose
        if error_format is None:
            error_format = ctx.obj.error_format
        if timestats is None:
            timestats = ctx.obj.timestats
        if pdb is None:
            pdb = ctx.obj.pdb

    if transpose and output_format not in TABLE_FORMATS:
        raise click_exception(
            "Transposing output tables (-x / --transpose) conflicts with "
            "non-table output format (-o / --output-format): {of}".
            format(of=output_format),
            error_format)

    if no_verify and ca_certs:
        raise click_exception(
            "Disabling HMC certificate verification (-n / --no-verify / "
            "ZHMC_NO_VERIFY) conflicts with specifying a CA certificate path "
            "(-c / --ca-certs / ZHMC_CA_CERTS)",
            error_format)

    # TODO: Add context support for the following options:
    if log is None:
        log = DEFAULT_LOG
    if log_dest is None:
        log_dest = DEFAULT_LOG_DESTINATION
    if syslog_facility is None:
        syslog_facility = DEFAULT_SYSLOG_FACILITY

    # Now we have the effective values for the options as they should be used
    # by the current command, regardless of the mode.

    # Set up logging

    for lc in LOG_COMPONENTS:
        reset_logger(lc)

    if log_dest == 'none':
        handler = None
    elif log_dest == 'syslog':
        # The choices in SYSLOG_FACILITIES have been validated by click
        # so we don't need to further check them.
        facility = SysLogHandler.facility_names[syslog_facility]
        system = platform.system()
        if system.startswith('CYGWIN_NT'):
            # Value is 'CYGWIN_NT-6.1'; strip off trailing version:
            system = 'CYGWIN_NT'
        try:
            addresses = SYSLOG_ADDRESSES[system]
        except KeyError:
            raise NotImplementedError(
                "Logging to syslog is not supported on this platform: {p}".
                format(p=system))
        assert isinstance(addresses, list)
        for address in addresses:
            try:
                handler = SysLogHandler(address=address, facility=facility)
            except Exception:  # nosec: B112 pylint: disable=broad-except
                continue
            break
        else:
            exc = sys.exc_info()[1]
            exc_name = exc.__class__.__name__ if exc else None
            raise RuntimeError(
                "Creating SysLogHandler with addresses {al!r} failed. "
                "Failure on last address {a!r} was: {exc}: {msg}".
                format(al=addresses, a=address, exc=exc_name, msg=exc))
        fs = '%(levelname)s %(name)s: %(message)s'
        handler.setFormatter(logging.Formatter(fs))
    elif log_dest == 'stderr':
        handler = StreamHandler(stream=sys.stderr)
        fs = '%(levelname)s %(name)s: %(message)s'
        handler.setFormatter(logging.Formatter(fs))
    else:
        # log_dest is the path name of the log file
        try:
            handler = logging.FileHandler(log_dest)
        except OSError as exc:
            raise click_exception(
                "Cannot log to file {fn}: {exc}: {msg}".
                format(fn=log_dest, exc=exc.__class__.__name__, msg=exc),
                error_format)
        fs = '%(levelname)s %(name)s: %(message)s'
        handler.setFormatter(logging.Formatter(fs))

    log_specs = log.split(',')
    for log_spec in log_specs:

        # ignore extra ',' at begin, end or in between
        if log_spec == '':
            continue

        try:
            log_comp, log_level = log_spec.split('=', 1)
        except ValueError:
            raise click_exception("Missing '=' in COMP=LEVEL specification in "
                                  "--log option: {ls}".format(ls=log_spec),
                                  error_format)

        level = getattr(logging, log_level.upper(), None)
        if level is None:
            raise click_exception("Invalid log level in COMP=LEVEL "
                                  "specification in --log option: {ls}".
                                  format(ls=log_spec),
                                  error_format)

        if log_comp not in LOG_COMPONENTS:
            raise click_exception("Invalid log component in COMP=LEVEL "
                                  "specification in --log option: {ls}".
                                  format(ls=log_spec), error_format)

        if handler:
            setup_logger(log_comp, handler, level)

        logger = logging.getLogger('zhmcclient.hmc')
        logger.debug("zhmc logon source: %s", logon_source)

    if session_id and session_id.startswith('faked_session:'):
        # This should be used by the zhmc function tests only.
        # A SyntaxError raised by an incorrect expression is considered
        # an internal error in the function tests and is therefore not
        # handled.
        expr = session_id.split(':', 1)[1]
        faked_session = eval(expr)  # nosec: B307 pylint: disable=eval-used
        assert isinstance(faked_session, zhmcclient_mock.FakedSession)
        session_id = faked_session

    def get_password_via_prompt(host, userid):
        """
        Password retrieval function that prompts for the password.

        It follows the interface defined in
        :func:`~zhmcclient.get_password_interface` and needs access to the
        click context (ctx).
        """
        if userid is not None and host is not None:
            ctx.obj.spinner.stop()
            password = get_password(host, userid)
            ctx.obj.spinner.start()
            return password

        raise click_exception("{cmd} command requires logon, but no "
                              "session-id or userid provided.".
                              format(cmd=ctx.invoked_subcommand),
                              error_format)

    # Make sure that in interactive mode, we have the password (otherwise,
    # every interactive command prompts for it).
    if password is None and ctx.invoked_subcommand in (None, 'repl'):
        password = get_password(host, userid)

    # We create a command context for each command: An interactive command has
    # its own command context different from the command context for the
    # command line.
    ctx.obj = CmdContext(
        params=ctx.params, host=host, userid=userid, password=password,
        no_verify=no_verify, ca_certs=ca_certs, logon_source=logon_source,
        session_name=session_name, output_format=output_format,
        transpose=transpose, error_format=error_format, timestats=timestats,
        session_id=session_id, get_password=get_password_via_prompt, pdb=pdb)

    # Invoke default command
    if ctx.invoked_subcommand is None:
        ctx.invoke(repl)


def get_password(host, userid):
    """
    Prompt for a password and return the password.
    """
    password = click.prompt(
        "Enter password (for user {userid} at HMC {host})".
        format(userid=userid, host=host), hide_input=True,
        confirmation_prompt=False, type=str, err=True)
    return password


def reset_logger(log_comp):
    """
    Reset the logger for the specified log component to have exactly one
    NullHandler.
    """

    name = LOGGER_NAMES[log_comp]
    logger = logging.getLogger(name)

    has_nh = False
    for h in logger.handlers:
        if not has_nh and isinstance(h, NullHandler):
            has_nh = True
            continue
        logger.removeHandler(h)

    if not has_nh:
        nh = NullHandler()
        logger.addHandler(nh)


def setup_logger(log_comp, handler, level):
    """
    Setup the logger for the specified log component to add the specified
    handler and to set it to the specified log level.
    """

    name = LOGGER_NAMES[log_comp]
    logger = logging.getLogger(name)

    logger.addHandler(handler)
    logger.setLevel(level)


@cli.command('help')
@click.pass_context
def repl_help(ctx):
    # pylint: disable=unused-argument
    """
    Show help message for interactive mode.

    Parameters:

      ctx (:class:`click.Context`): The click context object. Created by the
        ``@click.pass_context`` decorator.
    """
    print("""
The following can be entered in interactive mode:

  <zhmc-cmd>                  Execute zhmc command <zhmc-cmd>.
  !<shell-cmd>                Execute shell command <shell-cmd>.

  <CTRL-D>, :q, :quit, :exit  Exit interactive mode.

  <TAB>                       Tab completion (can be used anywhere).
  --help                      Show zhmc general help message, including a list
                              of zhmc commands.
  <zhmc-cmd> --help           Show help message for zhmc command <zhmc-cmd>.
  help                        Show this help message.
  :?, :h, :help               Show (incomplete) help message about interactive
                              mode.
""")


@cli.command('repl')
@click.pass_context
def repl(ctx):
    """
    Enter interactive (REPL) mode (default).

    Parameters:

      ctx (:class:`click.Context`): The click context object. Created by the
        ``@click.pass_context`` decorator.
    """

    history_file = REPL_HISTORY_FILE
    if history_file.startswith('~'):
        history_file = os.path.expanduser(history_file)

    print("Enter 'help' for help, <CTRL-D> or ':q' to exit.")

    prompt_kwargs = {
        'message': REPL_PROMPT,
        'history': FileHistory(history_file),
    }
    click_repl.repl(ctx, prompt_kwargs=prompt_kwargs)


# TODO: Apparently registering is not needed, clarify that.
# click_repl.register_repl(repl)



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HMC password rule management.
"""


import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props


def find_password_rule(cmd_ctx, console, password_rule_name):
    """
    Find a password rule by name and return its resource object.
    """
    try:
        password_rule = console.password_rules.find(name=password_rule_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return password_rule


@cli.group('passwordrule', options_metavar=COMMAND_OPTIONS_METAVAR)
def password_rule_group():
    """
    Command group for managing HMC password rules.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@password_rule_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def password_rule_list(cmd_ctx, **options):
    """
    List the HMC password rules.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_password_rule_list(cmd_ctx, options))


@password_rule_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.pass_obj
def password_rule_show(cmd_ctx, password_rule):
    """
    Show the details of an HMC password rule.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Console.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_password_rule_show(cmd_ctx, password_rule))


@password_rule_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, required=True,
              help='The name of the new password rule.')
@click.option('--description', type=str, required=False,
              help='The description of the new password rule.')
@click.option('--expiration', type=int, required=False,
              help='The total number of days a password is valid before '
              'it expires. '
              '0 indicates that the password never expires. '
              'Default: 0')
@click.option('--min-length', type=int, required=False,
              help='The minimum required length of the password. '
              'Default: 8')
@click.option('--max-length', type=int, required=False,
              help='The maximum allowed length of the password. '
              'Default: 256')
@click.option('--consecutive-characters', type=int, required=False,
              help='The maximum number of characters that are allowed to '
              'be repeated in a row. '
              '0 indicates that there is no such limit. '
              'Default: 0')
@click.option('--similarity-count', type=int, required=False,
              help='The maximum number of consecutive characters in the '
              'current password that can match consecutive characters in '
              'the previous password. '
              '0 indicates that there is no such limit. '
              'Default: 0')
@click.option('--history-count', type=int, required=False,
              help='The number of previous passwords to which a new '
              'password is compared for uniqueness. '
              '0 indicates that there is no such comparison. '
              'Default: 0')
@click.option('--case-sensitive', type=bool, required=False,
              help='The indicator whether the password is case sensitive. '
              'Default: False')
@click.pass_obj
def password_rule_create(cmd_ctx, **options):
    """
    Create a user-defined HMC password rule.

    Note that the characters rules of password rules can be managed via the
    'characterrule' command group.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_password_rule_create(cmd_ctx, options))


@password_rule_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.option('--description', type=str, required=False,
              help='The new description of the password rule.')
@click.option('--expiration', type=int, required=False,
              help='The new total number of days a password is valid before '
              'it expires. '
              '0 indicates that the password never expires.')
@click.option('--min-length', type=int, required=False,
              help='The new minimum required length of the password.')
@click.option('--max-length', type=int, required=False,
              help='The new maximum allowed length of the password.')
@click.option('--consecutive-characters', type=int, required=False,
              help='The new maximum number of characters that are allowed to '
              'be repeated in a row. '
              '0 indicates that there is no such limit.')
@click.option('--similarity-count', type=int, required=False,
              help='The new maximum number of consecutive characters in the '
              'current password that can match consecutive characters in '
              'the previous password. '
              '0 indicates that there is no such limit.')
@click.option('--history-count', type=int, required=False,
              help='The new number of previous passwords to which a new '
              'password is compared for uniqueness. '
              '0 indicates that there is no such comparison.')
@click.option('--case-sensitive', type=int, required=False,
              help='The new indicator whether the password is case sensitive.')
@click.pass_obj
def password_rule_update(cmd_ctx, password_rule, **options):
    """
    Update the properties of an HMC password rule.

    Note that HMC password rules cannot be renamed.

    Note that the characters rules of password rules can be managed via the
    'characterrule' command group.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_password_rule_update(
        cmd_ctx, password_rule, options))


@password_rule_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the password rule.',
              prompt='Are you sure you want to delete this password rule ?')
@click.pass_obj
def password_rule_delete(cmd_ctx, password_rule):
    """
    Delete a user-defined HMC password rule.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_password_rule_delete(
        cmd_ctx, password_rule))


def cmd_password_rule_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    additions = {}

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    try:
        password_rules = console.password_rules.list(
            full_properties=False, filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, password_rules, cmd_ctx.output_format,
                        show_list, additions, all=options['all'],
                        sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_password_rule_show(cmd_ctx, password_rule_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)

    try:
        password_rule.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(password_rule.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = console.name

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_password_rule_create(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    org_options = original_options(options)
    properties = options_to_properties(org_options, None)

    try:
        new_password_rule = console.password_rules.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New password rule '{r}' has been created.".
               format(r=new_password_rule.properties['name']))


def cmd_password_rule_update(cmd_ctx, password_rule_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)

    org_options = original_options(options)
    properties = options_to_properties(org_options, None)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating password rule '{r}'.".
                   format(r=password_rule_name))
        return

    try:
        password_rule.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Password rule '{r}' has been updated.".
               format(r=password_rule_name))


def cmd_password_rule_delete(cmd_ctx, password_rule_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)

    try:
        password_rule.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Password rule '{r}' has been deleted.".
               format(r=password_rule_name))
